from typing import TypeVar

import numpy as np
import numpy.typing as npt

_T = TypeVar("_T", bound=npt.NBitBase)


def make_rotation_matrix(
    z1: "npt.NDArray[np.floating[_T]]", z2: "npt.NDArray[np.floating[_T]]"
) -> "npt.NDArray[np.floating[_T]]":
    v0: "npt.NDArray[np.floating[_T]]" = z1 / np.linalg.norm(z1)
    v1: "npt.NDArray[np.floating[_T]]" = z2 / np.linalg.norm(z2)

    angle = np.arccos(np.clip(np.dot(v0, v1), -1.0, 1.0))
    direction = np.cross(v0, v1)

    sina = np.sin(angle)
    cosa = np.cos(angle)
    direction[:3] /= np.sqrt(np.dot(direction[:3], direction[:3]))
    # rotation matrix around unit vector
    rotation_matrix = np.diag([cosa, cosa, cosa])
    rotation_matrix += np.outer(direction, direction) * (1.0 - cosa)
    direction *= sina
    rotation_matrix += np.array(
        [
            [0.0, -direction[2], direction[1]],
            [direction[2], 0.0, -direction[0]],
            [-direction[1], direction[0], 0.0],
        ]
    )
    final_rotation_matrix = np.identity(4, dtype=z1.dtype)
    final_rotation_matrix[:3, :3] = rotation_matrix

    return final_rotation_matrix


def make_scale_matrix(factor: float) -> npt.NDArray[np.float32]:
    return np.diag([factor, factor, factor, 1.0])


def make_translation_matrix(
    direction: "npt.NDArray[np.floating[_T]]",
) -> "npt.NDArray[np.floating[_T]]":
    translation_matrix = np.identity(4, dtype=direction.dtype)
    translation_matrix[:3, 3] = direction[:3]
    return translation_matrix



import numpy as np
from numba import jit  # type: ignore [attr-defined]
from numba import njit


@njit(
    "boolean(float32[:,:], float32[:], float32)", fastmath=True, nogil=True, cache=True
)
def is_point_far_enough(points, tested_point, squared_min_distance):
    nbp = points.shape[0]
    farenough = True
    for i in range(nbp - 1, -1, -1):
        if (
            (tested_point[0] - points[i][0]) ** 2
            + (tested_point[1] - points[i][1]) ** 2
            + (tested_point[2] - points[i][2]) ** 2
        ) < squared_min_distance:
            farenough = False
            break
    return farenough


@jit(cache=True, nogil=True, nopython=True)
def xyz_to_child_index(xyz, aabb_center):
    test = np.greater_equal(xyz - aabb_center, 0).astype(np.int8)
    return np.sum(np.left_shift(test, np.array([2, 1, 0])), axis=1)


@njit(
    "int32[:](float32[:,:], int32[:], float32[:], float32[:], int32)",
    cache=True,
    nogil=True,
)
def xyz_to_key(xyz, cell_count, aabb_min, inv_aabb_size, shift):
    a = ((cell_count * inv_aabb_size) * (xyz - aabb_min)).astype(np.int64)
    a = np.minimum(np.maximum(a, 0), cell_count - 1)
    a[:, 1] <<= shift
    a[:, 2] <<= 2 * shift
    return np.sum(a, axis=1).astype(np.int32)



import argparse
import copy
from pathlib import Path
from typing import Any, Dict, List, Optional, TypeVar

import numpy as np
import numpy.typing as npt

from py3dtiles.exceptions import (
    BoundingVolumeMissingException,
    InvalidTilesetError,
    TilerException,
)
from py3dtiles.tileset.bounding_volume_box import BoundingVolumeBox
from py3dtiles.tileset.content import Pnts
from py3dtiles.tileset.tile import Tile
from py3dtiles.tileset.tileset import TileSet
from py3dtiles.utils import split_aabb

_T = TypeVar("_T", bound=npt.NBitBase)


def merge(
    tilesets: List[TileSet], tileset_paths: Optional[Dict[TileSet, Path]] = None
) -> TileSet:
    """
    Create a tileset that include all input tilesets. The tilesets don't need to be written.
    The output tileset is not written but return as dict (TilesetDictType).
    """
    if not tilesets:
        raise ValueError("The tileset list cannot be empty")

    global_tileset = TileSet()
    for tileset in tilesets:
        bounding_volume = copy.deepcopy(tileset.root_tile.bounding_volume)
        if bounding_volume is None:
            raise BoundingVolumeMissingException(
                "The root tile of all tilesets should have a bounding volume"
            )

        bounding_volume.transform(tileset.root_tile.transform)

        tile = Tile(
            geometric_error=tileset.root_tile.geometric_error,
            bounding_volume=bounding_volume,
            refine_mode="REPLACE",
        )
        if tileset_paths is not None:
            tile.content_uri = tileset_paths[tileset]
        else:
            tile.tile_content = tileset

        global_tileset.root_tile.add_child(tile)

    biggest_geometric_error = 0.0
    for child in global_tileset.root_tile.children:
        biggest_geometric_error = max(biggest_geometric_error, child.geometric_error)

    global_tileset.geometric_error = biggest_geometric_error
    global_tileset.root_tile.geometric_error = biggest_geometric_error
    global_tileset.root_tile.set_refine_mode("REPLACE")

    return global_tileset


def quadtree_split(
    aabb: "npt.NDArray[np.floating[_T]]",
) -> List["npt.NDArray[np.floating[_T]]"]:
    return [
        split_aabb(aabb, 0, True),
        split_aabb(aabb, 2, True),
        split_aabb(aabb, 4, True),
        split_aabb(aabb, 6, True),
    ]


def is_point_inside(
    point: "npt.NDArray[np.floating[_T]]", aabb: "npt.NDArray[np.floating[_T]]"
) -> np.bool_:
    return np.all(aabb[0] <= point) and np.all(point < aabb[1])


def build_tileset_quadtree(
    aabb: npt.NDArray[np.float64],
    tilesets: List[TileSet],
    bounding_box_centers: List[npt.NDArray[np.float64]],
    inv_base_transform: npt.NDArray[np.float64],
    tileset_paths: Optional[Dict[TileSet, Path]] = None,
) -> Optional[Tile]:
    insides = [
        (tileset, center)
        for tileset, center in zip(tilesets, bounding_box_centers)
        if is_point_inside(center, aabb)
    ]

    if len(insides) == 0:
        return None

    tileset_insides = [tileset for tileset, _ in insides]
    center_insides = [center for _, center in insides]

    quadtree_diag = np.linalg.norm(aabb[1][:2] - aabb[0][:2])

    if len(tileset_insides) == 1 or quadtree_diag < 1:
        # apply transform to boundingVolume
        tileset = tileset_insides[0]
        bvb = copy.deepcopy(tileset.root_tile.bounding_volume)
        if bvb is None:
            raise InvalidTilesetError(
                "The root tile of the tileset must have a bounding volume."
            )
        if tileset.root_tile.transform is not None:
            bvb.transform(tileset.root_tile.transform)

        tile = Tile(
            geometric_error=tileset.root_tile.geometric_error,
            transform=inv_base_transform,
            bounding_volume=bvb,
        )
        if tileset_paths is not None:
            tile.content_uri = tileset_paths[tileset]
        else:
            tile.tile_content = tileset

        return tile
    else:
        children = []

        for quarter in quadtree_split(aabb):
            r = build_tileset_quadtree(
                quarter,
                tileset_insides,
                center_insides,
                inv_base_transform,
                tileset_paths,
            )
            if r is not None:
                children.append(r)

        main_root_tile = tileset_insides[0].root_tile
        union_aabb = copy.deepcopy(main_root_tile.bounding_volume)
        if union_aabb is None:
            raise InvalidTilesetError(
                "The root tile of the tileset must have a bounding volume."
            )

        if main_root_tile.transform is not None:
            union_aabb.transform(main_root_tile.transform)

        # take half points from our children
        xyz = np.zeros((0, 3), dtype=np.float32)
        rgb = np.zeros((0, 3), dtype=np.uint8)

        max_point_count = 50000
        point_count = 0
        for tileset in tileset_insides:
            if tileset.root_tile.tile_content is not None:
                root_tile_content = tileset.root_tile.tile_content
            elif tileset_paths is not None:
                root_tile_content = tileset.root_tile.get_or_fetch_content(
                    tileset_paths[tileset]
                )
            else:
                root_tile_content = None

            if root_tile_content is not None:
                if not isinstance(root_tile_content, Pnts):
                    raise TilerException(
                        "The tileset must only have Pnts as tile content on the root tile."
                    )
                point_count += root_tile_content.body.feature_table.header.points_length

        ratio = min(0.5, max_point_count / point_count)

        for tileset in tileset_insides:
            root_tile = tileset.root_tile
            root_tile_content = root_tile.tile_content
            if not isinstance(root_tile_content, Pnts):
                raise TilerException(
                    "The tileset must only have Pnts as tile content on the root tile."
                )

            local_transform = root_tile.transform @ inv_base_transform
            _xyz, _rgb = root_tile_content.body.get_points(local_transform)

            select = np.random.choice(_xyz.shape[0], int(_xyz.shape[0] * ratio))
            xyz = np.concatenate((xyz, _xyz[select]))
            if _rgb is not None:
                rgb = np.concatenate((rgb, _rgb[select]))

            ab = copy.deepcopy(root_tile.bounding_volume)
            if ab is None:
                raise InvalidTilesetError(
                    "The root tile of the tileset must have a bounding volume."
                )

            if root_tile.transform is not None:
                ab.transform(root_tile.transform)
            union_aabb.add(ab)

        pnts = Pnts.from_points(
            np.concatenate((xyz.view(np.uint8).ravel(), rgb.ravel())),
            rgb.shape[0] > 0,
            False,  # TODO: Handle classification in the merging process
            False,  # TODO: Handle intensity in the merging process
        )

        union_aabb.transform(inv_base_transform)

        tile = Tile(
            refine_mode="REPLACE",
            geometric_error=max(
                [tileset.root_tile.geometric_error for tileset in tileset_insides]
            )
            / ratio,
        )
        if pnts is not None:
            tile.tile_content = pnts
            tile.content_uri = Path("r.pnts")

        for child in children:
            tile.add_child(child)

        return tile


def merge_with_pnts_content(
    tilesets: List[TileSet], tileset_paths: Optional[Dict[TileSet, Path]] = None
) -> TileSet:
    global_bounding_volume = BoundingVolumeBox()
    bounding_box_centers = []

    for tileset in tilesets:
        # apply transformation
        if tileset.root_tile.bounding_volume is None:
            raise BoundingVolumeMissingException(
                "The root tile should have a bounding volume."
            )

        bounding_box = copy.deepcopy(tileset.root_tile.bounding_volume)
        if tileset.root_tile.transform is not None:
            bounding_box.transform(tileset.root_tile.transform)

        global_bounding_volume.add(bounding_box)

        bounding_box_centers.append(bounding_box.get_center())

    corners = global_bounding_volume.get_corners()
    aabb = np.array((corners[0], corners[-1]))

    base_transform = tilesets[0].root_tile.transform
    inv_base_transform = np.linalg.inv(base_transform)

    # build hierarchical structure
    result = build_tileset_quadtree(
        aabb, tilesets, bounding_box_centers, inv_base_transform, tileset_paths
    )

    if result is None:
        raise RuntimeError("Result shouldn't be None")

    result.transform = base_transform
    tileset = TileSet(geometric_error=float(np.linalg.norm((aabb[1] - aabb[0])[0:3])))
    tileset.root_tile = result

    return tileset


def merge_from_files(
    tileset_paths: List[Path],
    output_tileset_path: Path,
    overwrite: bool = True,
    force_universal_merger: bool = False,
) -> None:
    output_tileset_path = output_tileset_path.absolute()
    if output_tileset_path.exists():
        if overwrite:
            TileSet.from_file(output_tileset_path).delete_on_disk(
                output_tileset_path, delete_sub_tileset=False
            )
        else:
            raise FileExistsError(
                f"Destination tileset {output_tileset_path} already exists."
            )

    tilesets = []
    for path in tileset_paths:
        tilesets.append(TileSet.from_file(path))

    not_only_pnts = force_universal_merger or any(
        not isinstance(
            tileset.root_tile.get_or_fetch_content(tileset_path.parent), Pnts
        )
        for tileset_path, tileset in zip(tileset_paths, tilesets)
    )

    relative_tileset_paths = {
        tileset: path.absolute().relative_to(output_tileset_path.parent)
        for tileset, path in zip(tilesets, tileset_paths)
    }

    if not_only_pnts:
        tileset = merge(tilesets, relative_tileset_paths)
    else:
        tileset = merge_with_pnts_content(tilesets, relative_tileset_paths)

    tileset.root_uri = output_tileset_path.parent
    tileset.write_to_directory(output_tileset_path)


def _init_parser(
    subparser: "argparse._SubParsersAction[Any]",
) -> argparse.ArgumentParser:
    parser: argparse.ArgumentParser = subparser.add_parser(
        "merge",
        help="Merge several pointcloud tilesets in 1 tileset. All input tilesets must be relative to the output tileset.",
    )
    parser.add_argument("tilesets", nargs="+", help="All tileset paths to merge")
    parser.add_argument(
        "--output-tileset", required=True, help="The path to the output tileset."
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite the output folder if it already exists.",
    )

    return parser


def _main(args: argparse.Namespace) -> None:
    return merge_from_files(
        [Path(tileset_file) for tileset_file in args.tilesets],
        Path(args.output_tileset),
        args.overwrite,
    )



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# -- Project information -----------------------------------------------------

project = "py3dtiles"
copyright = "2023, the py3dtiles contributors"  # noqa: A001
author = "Augustin Trancart, Paul Blottière, Jérémy Gaillard, Ludovic Delauné, Nicolas Saul, Pierre-Éric Pelloux-Prayer, Raphaël Delhome, Vincent Jaillot, Éric Lemoine, Antoine Facchini"

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
    "sphinx_multiversion",
    "sphinxcontrib.apidoc",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"display_version": True}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.

# -- Options for sphinxcontrib.apidoc
apidoc_module_dir = "../py3dtiles"
apidoc_output_dir = "./api"
apidoc_excluded_paths = ["tests"]
apidoc_separate_modules = True
apidoc_module_first = True
apidoc_toc_file = False
apidoc_extra_args = ["-d 3"]

# -- Options for sphinx-multiversion

# Whitelist pattern for tags (set to None to ignore all tags)
# KEEP THAT IN SYNC with the current_version calculation in the "pages" job in .gitlab-ci.yml
smv_tag_whitelist = r"^v(\d+\.)?(\d+\.)?(\*|\d+)$"

# Whitelist pattern for branches (set to None to ignore all branches)
smv_branch_whitelist = r"^main$"

# Whitelist pattern for remotes (set to None to use local branches only)
# set to upstream because we add this remote for gitlab, see .gitlab-ci.yml
smv_remote_whitelist = r"^upstream|origin$"

# Pattern for released versions
# KEEP THAT IN SYNC with the current_version calculation in the "pages" job in .gitlab-ci.yml
smv_released_pattern = r"^tags/^v(\d+\.)?(\d+\.)?(\*|\d+)$"

# Format for versioned output directories inside the build directory
smv_outputdir_format = "{ref.name}"

# Determines whether remote or local git branches/tags are preferred if their output dirs conflict
smv_prefer_remote_refs = True

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "py3dtilesdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "py3dtiles.tex",
        "py3dtiles Documentation",
        "Augustin Trancart, Paul Blottière, Jérémy Gaillard, Ludovic Delauné, Nicolas Saul, Pierre-Éric Pelloux-Prayer, Raphaël Delhome, Vincent Jaillot, Éric Lemoine, Antoine Facchini",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "py3dtiles", "py3dtiles Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "py3dtiles",
        "py3dtiles Documentation",
        author,
        "py3dtiles",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for sphinx.ext.autosectionlabel ----------------------------------
autosectionlabel_prefix_document = True



from .base_extension import (
    BaseExtension,
    create_extension,
    is_extension_registered,
    register_extension,
)
from .batch_table_hierarchy_extension import BatchTableHierarchy

__all__ = [
    "BaseExtension",
    "BatchTableHierarchy",
    "create_extension",
    "is_extension_registered",
    "register_extension",
]



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    Union,
)

import numpy as np
import numpy.typing as npt
from pyproj import CRS

if TYPE_CHECKING:
    from typing_extensions import NotRequired

# Tileset types

ExtensionDictType = Dict[str, Any]
ExtraDictType = Dict[str, Any]
GeometricErrorType = float
PropertyType = Dict[str, Any]
RefineType = Literal["ADD", "REPLACE"]
TransformDictType = List[float]


class RootPropertyDictType(TypedDict):
    extensions: NotRequired[dict[str, ExtensionDictType]]
    extras: NotRequired[ExtraDictType]


class BoundingVolumeBoxDictType(RootPropertyDictType):
    box: list[float]


class BoundingVolumeRegionDictType(RootPropertyDictType):
    region: list[float]


class BoundingVolumeSphereDictType(RootPropertyDictType):
    sphere: list[float]


BoundingVolumeDictType = Union[
    BoundingVolumeBoxDictType,
    BoundingVolumeRegionDictType,
    BoundingVolumeSphereDictType,
]


class ContentType(RootPropertyDictType):
    boundingVolume: NotRequired[BoundingVolumeDictType]
    uri: str


class PropertyDictType(RootPropertyDictType):
    maximum: float
    minimum: float


class AssetDictType(RootPropertyDictType):
    version: Literal["1.0", "1.1"]
    tilesetVersion: NotRequired[str]


class TileDictType(RootPropertyDictType):
    boundingVolume: BoundingVolumeDictType
    geometricError: GeometricErrorType
    viewerRequestVolume: NotRequired[BoundingVolumeDictType]
    refine: NotRequired[RefineType]
    transform: NotRequired[TransformDictType]
    content: NotRequired[ContentType]
    children: NotRequired[list[TileDictType]]


class TilesetDictType(RootPropertyDictType):
    asset: AssetDictType
    geometricError: GeometricErrorType
    root: TileDictType
    properties: NotRequired[PropertyType]
    extensionsUsed: NotRequired[list[str]]
    extensionsRequired: NotRequired[list[str]]


# Tile content types

BatchTableHeaderDataType = Dict[str, Union[List[Any], Dict[str, Any]]]

FeatureTableHeaderDataType = Dict[
    str,
    Union[
        int,  # points_length
        Dict[str, int],  # byte offsets
        Tuple[float, float, float],  # rtc
        List[float],  # quantized_volume_offset and quantized_volume_scale
        Tuple[int, int, int, int],  # constant_rgba
    ],
]


class HierarchyClassDictType(TypedDict):
    name: str
    length: int
    instances: dict[str, list[Any]]


# Tiler types

PortionItemType = Tuple[int, ...]
PortionsType = List[Tuple[str, PortionItemType]]


class MetadataReaderType(TypedDict):
    portions: PortionsType
    aabb: npt.NDArray[np.float64]
    crs_in: CRS | None
    point_count: int
    avg_min: npt.NDArray[np.float64]


OffsetScaleType = Tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    Optional[npt.NDArray[np.float64]],
    Optional[float],
]



import argparse
from pathlib import Path
from typing import Any

from py3dtiles.tileset.content import read_binary_tile_content


def _main(args: argparse.Namespace) -> None:
    try:
        tile_content = read_binary_tile_content(args.file)
    except ValueError as e:
        print(f"Error when reading the file {args.file}")
        raise e

    print(f"=== Tile content: ===\n{tile_content}\n=====================")


def _init_parser(
    subparser: "argparse._SubParsersAction[Any]",
) -> argparse.ArgumentParser:
    # arg parse
    parser: argparse.ArgumentParser = subparser.add_parser(
        "info", help="Extract information from a 3DTiles file"
    )

    parser.add_argument("file", type=Path)

    return parser



from __future__ import annotations

import struct
from pathlib import Path

import numpy as np
import numpy.typing as npt

from py3dtiles.exceptions import InvalidPntsError

from .batch_table import BatchTable
from .pnts_feature_table import (
    PntsFeatureTable,
    PntsFeatureTableBody,
    PntsFeatureTableHeader,
    SemanticPoint,
)
from .tile_content import TileContent, TileContentBody, TileContentHeader


class Pnts(TileContent):
    def __init__(self, header: PntsHeader, body: PntsBody) -> None:
        super().__init__()
        self.header: PntsHeader = header
        self.body: PntsBody = body
        self.sync()

    def sync(self) -> None:
        """
        Synchronizes headers with the Pnts body.
        """
        self.header.ft_json_byte_length = len(self.body.feature_table.header.to_array())
        self.header.ft_bin_byte_length = len(self.body.feature_table.body.to_array())
        self.header.bt_json_byte_length = len(self.body.batch_table.header.to_array())
        self.header.bt_bin_byte_length = len(self.body.batch_table.body.to_array())

        self.header.tile_byte_length = (
            PntsHeader.BYTE_LENGTH
            + self.header.ft_json_byte_length
            + self.header.ft_bin_byte_length
            + self.header.bt_json_byte_length
            + self.header.bt_bin_byte_length
        )

    @staticmethod
    def from_features(
        feature_table_header: PntsFeatureTableHeader,
        position_array: npt.NDArray[np.float32 | np.uint16],
        color_array: npt.NDArray[np.uint8 | np.uint16] | None = None,
        normal_position: npt.NDArray[np.float32 | np.uint8] | None = None,
    ) -> Pnts:
        """
        Creates a Pnts from features defined by pd_type and cd_type.
        """
        pnts_body = PntsBody()
        pnts_body.feature_table = PntsFeatureTable.from_features(
            feature_table_header, position_array, color_array, normal_position
        )

        pnts = Pnts(PntsHeader(), pnts_body)
        pnts.sync()

        return pnts

    @staticmethod
    def from_array(array: npt.NDArray[np.uint8]) -> Pnts:
        """
        Creates a Pnts from an array
        """

        # build tile header
        h_arr = array[0 : PntsHeader.BYTE_LENGTH]
        pnts_header = PntsHeader.from_array(h_arr)

        if pnts_header.tile_byte_length != len(array):
            raise InvalidPntsError(
                f"Invalid byte length in header, the size of array is {len(array)}, "
                f"the tile_byte_length for header is {pnts_header.tile_byte_length}"
            )

        # build tile body
        b_len = (
            pnts_header.ft_json_byte_length
            + pnts_header.ft_bin_byte_length
            + pnts_header.bt_json_byte_length
            + pnts_header.bt_bin_byte_length
        )
        b_arr = array[PntsHeader.BYTE_LENGTH : PntsHeader.BYTE_LENGTH + b_len]
        pnts_body = PntsBody.from_array(pnts_header, b_arr)

        # build the tile with header and body
        return Pnts(pnts_header, pnts_body)

    @staticmethod
    def from_points(
        points: npt.NDArray[np.uint8],
        include_rgb: bool,
        include_classification: bool,
        include_intensity: bool,
    ) -> Pnts:
        """
        Create a pnts from an uint8 data array containing:

        - points as SemanticPoint.POSITION
        - if include_rgb, rgb as SemanticPoint.RGB
        - if include_classification, classification as a single np.uint8 value that will be put in the batch table
        - if include_intensity, intensity as a single np.uint8 value that will be put in the batch table

        :param include_rgb: Whether the points array contains rgb values
        :param include_classification: Whether the point array contains classification values
        :param include_intensity: whether the point array contains intensity values
        :param points: the points array. Contains at least 3
        """
        if len(points) == 0:
            raise ValueError("The argument points cannot be empty.")

        point_size = (
            3 * 4
            + (3 if include_rgb else 0)
            + (1 if include_classification else 0)
            + (1 if include_intensity else 0)
        )

        if len(points) % point_size != 0:
            raise ValueError(
                f"The length of points array is {len(points)} but the point size is {point_size}."
                f"There is a rest of {len(points) % point_size}"
            )

        count = len(points) // point_size

        ft = PntsFeatureTable()
        ft.header = PntsFeatureTableHeader.from_semantic(
            SemanticPoint.POSITION,
            SemanticPoint.RGB if include_rgb else None,
            None,
            count,
        )
        ft.body = PntsFeatureTableBody.from_array(ft.header, points)

        bt = BatchTable()
        if include_classification:
            sdt = np.dtype([("Classification", "u1")])
            offset = count * (3 * 4 + (3 if include_rgb else 0))
            bt.add_property_as_binary(
                "Classification",
                points[offset : offset + count * sdt.itemsize],
                "UNSIGNED_BYTE",
                "SCALAR",
            )

        if include_intensity:
            sdt = np.dtype([("Intensity", "u1")])
            offset = count * (
                3 * 4 + (3 if include_rgb else 0) + (1 if include_classification else 0)
            )
            bt.add_property_as_binary(
                "Intensity",
                points[offset : offset + count * sdt.itemsize],
                "UNSIGNED_BYTE",
                "SCALAR",
            )

        body = PntsBody()
        body.feature_table = ft
        body.batch_table = bt

        pnts = Pnts(PntsHeader(), body)
        pnts.sync()

        return pnts

    @staticmethod
    def from_file(tile_path: Path) -> Pnts:
        with tile_path.open("rb") as f:
            data = f.read()
            arr = np.frombuffer(data, dtype=np.uint8)
            return Pnts.from_array(arr)


class PntsHeader(TileContentHeader):
    BYTE_LENGTH = 28

    def __init__(self) -> None:
        super().__init__()
        self.magic_value = b"pnts"
        self.version = 1

    def to_array(self) -> npt.NDArray[np.uint8]:
        """
        Returns the header as a numpy array.
        """
        header_arr = np.frombuffer(self.magic_value, np.uint8)

        header_arr2 = np.array(
            [
                self.version,
                self.tile_byte_length,
                self.ft_json_byte_length,
                self.ft_bin_byte_length,
                self.bt_json_byte_length,
                self.bt_bin_byte_length,
            ],
            dtype=np.uint32,
        )

        return np.concatenate((header_arr, header_arr2.view(np.uint8)))

    @staticmethod
    def from_array(array: npt.NDArray[np.uint8]) -> PntsHeader:
        """
        Create a PntsHeader from an array
        """

        h = PntsHeader()

        if len(array) != PntsHeader.BYTE_LENGTH:
            raise InvalidPntsError(
                f"Invalid header byte length, the size of array is {len(array)}, "
                f"the header must have a size of {PntsHeader.BYTE_LENGTH}"
            )

        h.version = struct.unpack("i", array[4:8].tobytes())[0]
        h.tile_byte_length = struct.unpack("i", array[8:12].tobytes())[0]
        h.ft_json_byte_length = struct.unpack("i", array[12:16].tobytes())[0]
        h.ft_bin_byte_length = struct.unpack("i", array[16:20].tobytes())[0]
        h.bt_json_byte_length = struct.unpack("i", array[20:24].tobytes())[0]
        h.bt_bin_byte_length = struct.unpack("i", array[24:28].tobytes())[0]

        return h


class PntsBody(TileContentBody):
    def __init__(self) -> None:
        self.feature_table: PntsFeatureTable = PntsFeatureTable()
        self.batch_table = BatchTable()

    def __str__(self) -> str:
        infos = {
            "feature_table_header": self.feature_table.header.to_json(),
            "points_length": self.feature_table.header.points_length,
        }
        if self.feature_table.header.points_length > 0:
            (
                feature_position,
                feature_color,
                feature_normal,
            ) = self.feature_table.get_feature_at(0)
            infos["first_point_position"] = feature_position
            infos["first_point_color"] = feature_color
            infos["first_point_normal"] = feature_normal
        infos["batch_table_header"] = self.batch_table.header.data
        for f in self.batch_table.header.data.keys():
            infos[f"- first point {f}"] = self.batch_table.get_binary_property(f)[0]
        return "\n".join(f"{key}: {value}" for key, value in infos.items())

    def to_array(self) -> npt.NDArray[np.uint8]:
        """
        Returns the body as a numpy array.
        """
        feature_table_array = self.feature_table.to_array()
        batch_table_array = self.batch_table.to_array()
        return np.concatenate((feature_table_array, batch_table_array))

    def get_points(
        self, transform: npt.NDArray[np.float64] | None
    ) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.uint8 | np.uint16] | None]:
        fth = self.feature_table.header

        xyz = self.feature_table.body.position.view(np.float32).reshape(
            (fth.points_length, 3)
        )
        if fth.colors == SemanticPoint.RGB:
            rgb = self.feature_table.body.color
            if rgb is None:
                raise InvalidPntsError(
                    "If fth.colors is SemanticPoint.RGB, rgb cannot be None."
                )
            rgb = rgb.reshape((fth.points_length, 3))
        else:
            rgb = None

        if transform is not None:
            transform = transform.reshape((4, 4), order="F")
            xyzw = np.hstack((xyz, np.ones((xyz.shape[0], 1), dtype=xyz.dtype)))
            xyz = np.dot(xyzw, transform.astype(xyz.dtype))[:, :3]

        return xyz, rgb

    @staticmethod
    def from_array(header: PntsHeader, array: npt.NDArray[np.uint8]) -> PntsBody:
        """
        Creates a PntsBody from an array and the header
        """

        # build feature table
        feature_table_size = header.ft_json_byte_length + header.ft_bin_byte_length
        feature_table_array = array[:feature_table_size]
        feature_table = PntsFeatureTable.from_array(header, feature_table_array)

        # build batch table
        batch_table_size = header.bt_json_byte_length + header.bt_bin_byte_length
        batch_table_array = array[
            feature_table_size : feature_table_size + batch_table_size
        ]
        batch_table = BatchTable.from_array(
            header, batch_table_array, feature_table.nb_points()
        )

        # build tile body with feature table
        body = PntsBody()
        body.feature_table = feature_table
        body.batch_table = batch_table

        return body



"""Test the point cloud readers.

The example that is run in the test (`simple.ply`) comes from the [CGAL repository](https://github.com/CGAL/cgal/blob/master/Data/data/points_3/b9_training.ply). Thanks to their maintainers (for more details, please refer to CGAL, Computational Geometry Algorithms Library, https://www.cgal.org):

"""

from pathlib import Path
from typing import Any, Dict

import numpy as np
import plyfile
from pytest import raises

from py3dtiles.reader import ply_reader


def test_ply_get_metadata(ply_filepath: Path) -> None:
    ply_metadata = ply_reader.get_metadata(path=ply_filepath)
    expected_point_count = 22300
    expected_aabb = (
        np.array([5.966480625e05, 2.43620015625e05, 7.350153350830078e01]),
        np.array([5.967389375e05, 2.43731984375e05, 9.718580627441406e01]),
    )
    assert list(ply_metadata.keys()) == [
        "portions",
        "aabb",
        "crs_in",
        "point_count",
        "avg_min",
    ]
    assert ply_metadata["portions"] == [(str(ply_filepath), (0, expected_point_count))]
    assert np.all(ply_metadata["aabb"][0] == expected_aabb[0])
    assert np.all(ply_metadata["aabb"][1] == expected_aabb[1])
    assert ply_metadata["crs_in"] is None
    assert ply_metadata["point_count"] == expected_point_count
    assert np.all(ply_metadata["avg_min"] == expected_aabb[0])


def test_ply_get_metadata_buggy(
    buggy_ply_data: Dict[str, Any], buggy_ply_filepath: Path
) -> None:
    buggy_ply_data["data"].write(buggy_ply_filepath)
    with raises(KeyError, match=buggy_ply_data["msg"]):
        _ = ply_reader.get_metadata(path=buggy_ply_filepath)
    buggy_ply_filepath.unlink()


def test_create_plydata_with_renamed_property(ply_filepath: Path) -> None:
    ply_data = plyfile.PlyData.read(ply_filepath)
    modified_ply_data = ply_reader.create_plydata_with_renamed_property(
        ply_data, "label", "classification"
    )
    for prop1, prop2 in zip(
        ply_data["vertex"].properties, modified_ply_data["vertex"].properties
    ):
        assert prop1.name == prop2.name or (
            prop1.name == "label" and prop2.name == "classification"
        )
    for dtype1, dtype2 in zip(
        ply_data["vertex"].data.dtype.names,
        modified_ply_data["vertex"].data.dtype.names,
    ):
        assert dtype1 == dtype2 or (dtype1 == "label" and dtype2 == "classification")



from __future__ import annotations

import json
from pathlib import Path

from py3dtiles.tileset import Tile, TileSet

DATA_DIRECTORY = Path(__file__).parent / "fixtures"


def test_constructor() -> None:
    tile_set = TileSet()
    assert tile_set.asset.to_dict() == {"version": "1.0"}
    assert tile_set.extensions == {}
    assert tile_set.geometric_error == 500
    assert isinstance(tile_set.root_tile, Tile)


def test_to_dict(tileset: TileSet) -> None:
    assert tileset.to_dict() == {
        "root": {
            "boundingVolume": {
                "box": [
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    9.0,
                    10.0,
                    11.0,
                    12.0,
                ]
            },
            "geometricError": 3.14159,
            "refine": "ADD",
        },
        "extensions": {"Test": {}},
        "extensionsUsed": ["Test"],
        "geometricError": 500,
        "asset": {"version": "1.0"},
    }


def test_from_dict(tmp_dir_with_content: Path) -> None:

    assert Path(tmp_dir_with_content, "tileset.json").exists()
    assert Path(tmp_dir_with_content, "r.pnts").exists()

    with (tmp_dir_with_content / "tileset.json").open() as f:
        tileset_dict = json.load(f)

    tileset = TileSet.from_dict(tileset_dict)
    tileset.root_uri = tmp_dir_with_content

    assert tileset.to_dict() == tileset_dict


def test_delete_on_disk(tileset_on_disk_with_sub_tileset_path: Path) -> None:
    # This test only checks if delete_on_disk doesn't delete sub-tileset

    tmp_folder = tileset_on_disk_with_sub_tileset_path.parent
    assert (tmp_folder / "tileset.json").exists()
    assert (tmp_folder / "upper_tileset.json").exists()

    tileset = TileSet.from_file(tileset_on_disk_with_sub_tileset_path)
    tileset.delete_on_disk(tmp_folder / "upper_tileset.json")

    assert (tmp_folder / "tileset.json").exists()
    assert (tmp_folder / "r.pnts").exists()
    assert not (tmp_folder / "upper_tileset.json").exists()


def test_delete_on_disk_with_sub_tileset(
    tileset_on_disk_with_sub_tileset_path: Path,
) -> None:
    # This test manly checks if delete_on_disk removes correctly all tile contents (binary and sub tileset)

    tmp_folder = tileset_on_disk_with_sub_tileset_path.parent
    assert (tmp_folder / "tileset.json").exists()
    assert (tmp_folder / "upper_tileset.json").exists()

    tileset = TileSet.from_file(tileset_on_disk_with_sub_tileset_path)
    tileset.delete_on_disk(tmp_folder / "upper_tileset.json", delete_sub_tileset=True)

    assert not (tmp_folder / "tileset.json").exists()
    assert not (tmp_folder / "r.pnts").exists()
    assert not (tmp_folder / "upper_tileset.json").exists()



# for all sort of size calculation, we consider a fake "thikness" of a point
# in most cases this is useless, but sometimes it's useful, for instance when
# calculating box sizes when all the points are coplanar
MIN_POINT_SIZE = 0.00001



from __future__ import annotations

import copy
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
import numpy.typing as npt

from py3dtiles.exceptions import InvalidTilesetError, Py3dtilesException, TilerException
from py3dtiles.typing import RefineType, TileDictType

from .bounding_volume import BoundingVolume
from .bounding_volume_box import BoundingVolumeBox
from .content import TileContent, read_binary_tile_content
from .root_property import RootProperty

if TYPE_CHECKING:
    from py3dtiles.tileset import TileSet

DEFAULT_TRANSFORMATION = np.identity(4, dtype=np.float64)
DEFAULT_TRANSFORMATION.setflags(write=False)

T = TypeVar("T", bound=np.generic)


class Tile(RootProperty[TileDictType]):
    def __init__(
        self,
        geometric_error: float = 500,
        bounding_volume: BoundingVolume[Any] | None = None,
        transform: npt.NDArray[np.float64] = DEFAULT_TRANSFORMATION,
        refine_mode: RefineType = "ADD",
        content_uri: Path | None = None,
    ) -> None:
        super().__init__()
        self.bounding_volume = bounding_volume
        self.geometric_error = geometric_error
        self._refine: RefineType = "ADD"
        self.set_refine_mode(refine_mode)
        self.tile_content: TileContent | TileSet | None = None
        self.content_uri: Path | None = content_uri
        self.children: list[Tile] = []
        # Some possible valid properties left un-delt with viewerRequestVolume
        self.transform = transform

    @classmethod
    def from_dict(cls, tile_dict: TileDictType) -> Tile:
        if "box" in tile_dict["boundingVolume"]:
            bounding_volume = BoundingVolumeBox.from_dict(tile_dict["boundingVolume"])  # type: ignore [arg-type]
        elif (
            "region" in tile_dict["boundingVolume"]
            or "sphere" in tile_dict["boundingVolume"]
        ):
            raise NotImplementedError(
                "The support of bounding volume region and sphere is not implemented yet"
            )
        else:
            raise InvalidTilesetError(
                f"The bounding volume {list(tile_dict['boundingVolume'].keys())[0]} is unknown"
            )

        tile = cls(
            geometric_error=tile_dict["geometricError"],
            bounding_volume=bounding_volume,
        )

        if "refine" in tile_dict:
            tile.set_refine_mode(tile_dict["refine"])

        if "transform" in tile_dict:
            tile.transform = np.array(tile_dict["transform"]).reshape((4, 4), order="F")

        if "children" in tile_dict:
            for child in tile_dict["children"]:
                tile.children.append(Tile.from_dict(child))

        if "content" in tile_dict:
            tile.content_uri = Path(tile_dict["content"]["uri"])

        tile.set_properties_from_dict(tile_dict)

        return tile

    def get_or_fetch_content(self, root_uri: Path | None) -> TileContent | TileSet:
        """
        If a `tile_content` content has been set, returns this content.
        If the tile content is None and a `tile_content` uri has been set, the tile will load the file and return its content.

        :param root_uri: the base uri which `tile.content_uri` is relative to. Usually the directory containing the tileset containing this tile.
        """
        self._load_tile_content(root_uri)
        if self.tile_content is None:
            raise Py3dtilesException(
                "self.tile_content cannot be None here, seems to be a py3dtiles issue."
            )
        return self.tile_content

    def has_content(self) -> bool:
        """
        Returns if there is a tile content (loaded or not).
        """
        return bool(self.tile_content is not None or self.content_uri)

    def has_content_loaded(self) -> bool:
        """
        Returns if there is a tile content loaded in this tile.
        """
        return self.tile_content is not None

    def delete_on_disk(self, root_uri: Path, delete_sub_tileset: bool = False) -> None:
        """Deletes all files linked to the tile and its children. The uri of the folder where tileset is, should be defined.

        :param root_uri: The folder where tileset is.
        :param delete_sub_tileset: If True, all tilesets present as tile content will be removed as well as their content. If False, the linked tilesets in tiles won't be removed.
        """
        for child in self.children:
            child.delete_on_disk(root_uri, delete_sub_tileset)

        # if there is no content_uri, there is no file to remove
        if self.content_uri is None:
            return

        if self.content_uri.is_absolute():
            tile_content_path = self.content_uri
        else:
            tile_content_path = root_uri / self.content_uri

        if tile_content_path.suffix == ".json":
            if delete_sub_tileset:
                self.get_or_fetch_content(root_uri).delete_on_disk(tile_content_path)  # type: ignore
        else:
            tile_content_path.unlink(missing_ok=True)

    def set_refine_mode(self, mode: RefineType) -> None:
        if mode != "ADD" and mode != "REPLACE":
            raise InvalidTilesetError(
                f"Unknown refinement mode {mode}. Should be either 'ADD' or 'REPLACE'."
            )
        self._refine = mode

    def get_refine_mode(self) -> RefineType:
        return self._refine

    def add_child(self, tile: Tile) -> None:
        self.children.append(tile)

        if tile.bounding_volume is not None:
            if self.bounding_volume is None:
                self.bounding_volume = copy.deepcopy(tile.bounding_volume)
                self.bounding_volume.transform(tile.transform)
            else:
                transformed_bounding_volume = copy.deepcopy(tile.bounding_volume)
                transformed_bounding_volume.transform(tile.transform)
                parent_inv_transform = np.linalg.inv(self.transform)
                transformed_bounding_volume.transform(parent_inv_transform)
                self.bounding_volume.add(transformed_bounding_volume)

    def get_all_children(self) -> list[Tile]:
        """
        :return: the recursive (across the children tree) list of the children
                 tiles
        """
        descendants = []
        for child in self.children:
            # Add the child...
            descendants.append(child)
            # and if (and only if) they are grand-children then recurse
            if child.children:
                descendants += child.get_all_children()
        return descendants

    def sync_bounding_volume_with_children(self) -> None:
        if self.bounding_volume is None:
            raise TilerException("This Tile has no bounding volume.")
        if not isinstance(self.bounding_volume, BoundingVolumeBox):
            raise NotImplementedError(
                "Don't know yet how to sync non box bounding volume."
            )

        # We consider that whatever information is present it is the
        # proper one (in other terms: when they are no sub-tiles this tile
        # is a leaf-tile and thus is has no synchronization to do)
        for child in self.children:
            child.sync_bounding_volume_with_children()

        # The information that depends on (is defined by) the children
        # nodes is limited to be bounding volume.
        self.bounding_volume.sync_with_children(self)

    def transform_coords(self, coords: npt.NDArray[T]) -> npt.NDArray[T]:
        """
        Transform coordinates according to this tile transformation property.

        NOTE: this only applies the current tile transformation, not the transformation of ancestors
        """
        assert coords.ndim == 2
        assert coords.shape[1] == 3
        new_coords = np.zeros(coords.shape, dtype=coords.dtype)

        for (i, coord) in enumerate(coords):
            new_coords[i] = self.transform_coord(coord)
        return new_coords

    def transform_coord(self, coord: npt.NDArray[T]) -> npt.NDArray[T]:
        """
        Transform one coordinate according to this tile transformation property.

        NOTE: this only applies the current tile transformation, not the transformation of ancestors
        """
        [x, y, z] = coord
        # code inspired from three.js
        e = self.transform.flatten(order="F")

        w = 1 / (e[3] * x + e[7] * y + e[11] * z + e[15])

        x = (e[0] * x + e[4] * y + e[8] * z + e[12]) * w
        y = (e[1] * x + e[5] * y + e[9] * z + e[13]) * w
        z = (e[2] * x + e[6] * y + e[10] * z + e[14]) * w

        return np.array([x, y, z])

    def write_content(self, root_uri: Path | None) -> None:
        """
        Write (or overwrite) the tile *content* to the directory specified
        as parameter and withing the relative filename designated by
        the tile's content uri.

        :param root_uri: the base uri of this tile, usually the folder where the tileset is
        """

        if self.tile_content is None:
            raise TilerException(
                "The tile has no tile content. "
                "A tile content should be added in the tile."
            )

        if self.content_uri is None:
            raise TilerException("tile.content_uri is None, cannot write tile content")

        if self.content_uri.is_absolute():
            content_path = self.content_uri
        else:
            if root_uri is None:
                raise ValueError(
                    "No root_uri given and tile.content_uri is not absolute"
                )
            content_path = root_uri / self.content_uri

        # Make sure the output directory exists (note that target_dir may
        # be a sub-directory of 'directory' because the uri might hold its
        # own path):
        content_path.parent.mkdir(parents=True, exist_ok=True)

        if isinstance(self.tile_content, TileContent):
            self.tile_content.save_as(content_path)
        else:
            self.tile_content.write_to_directory(content_path)

    def to_dict(self) -> TileDictType:
        if self.bounding_volume is not None:
            bounding_volume = self.bounding_volume
        else:
            raise InvalidTilesetError("Bounding volume is not set")
        bounding_volume_dict = bounding_volume.to_dict()

        refine = self._refine
        if refine not in ["ADD", "REPLACE"]:
            raise InvalidTilesetError(
                f"refine should be either ADD or REPLACE, currently {refine}."
            )

        # Mandatory items
        dict_data: TileDictType = {
            "boundingVolume": bounding_volume_dict,
            "geometricError": self.geometric_error,
            "refine": refine,
        }

        dict_data = self.add_root_properties_to_dict(dict_data)

        if not np.array_equal(self.transform, DEFAULT_TRANSFORMATION):
            dict_data["transform"] = self.transform.flatten("F").tolist()

        if self.children:
            # The children list exists indeed (for technical reasons) yet it
            # happens to be still empty. This would pollute the json output
            # by adding a "children" entry followed by an empty list. In such
            # case just remove that attributes entry:
            dict_data["children"] = [child.to_dict() for child in self.children]

        if self.content_uri:
            dict_data["content"] = {"uri": self.content_uri.as_posix()}
        return dict_data

    def _load_tile_content(self, root_uri: Path | None) -> None:
        if self.tile_content:
            return

        if not self.content_uri:
            raise RuntimeError("Cannot load a tile without a content_uri")

        if self.content_uri.is_absolute():
            uri = self.content_uri
        else:
            if root_uri is None:
                raise RuntimeError(
                    "Cannot load a tile without a root_uri if self.content_uri is relative"
                )
            uri = root_uri / self.content_uri

        if uri.suffix == ".json":
            from .tileset import TileSet

            self.tile_content = TileSet.from_file(uri)
        else:
            self.tile_content = read_binary_tile_content(uri)



from __future__ import annotations

from typing import Any, NamedTuple, cast

import numpy as np
import numpy.typing as npt
import pygltflib


class GltfAttribute(NamedTuple):
    name: str
    accessor_type: pygltflib.SCALAR | pygltflib.VEC2 | pygltflib.VEC3
    component_type: pygltflib.UNSIGNED_BYTE | pygltflib.UNSIGNED_INT | pygltflib.FLOAT
    array: npt.NDArray[np.uint8 | np.uint16 | np.uint32 | np.float32]


def get_component_type_from_dtype(dt: np.dtype[Any]) -> int:
    val = None
    if dt == np.int8:
        val = pygltflib.BYTE
    elif dt == np.uint8:
        val = pygltflib.UNSIGNED_BYTE
    elif dt == np.int16:
        val = pygltflib.SHORT
    elif dt == np.uint16:
        val = pygltflib.UNSIGNED_SHORT
    elif dt == np.uint32:
        val = pygltflib.UNSIGNED_INT
    elif dt == np.float32:
        val = pygltflib.FLOAT
    else:
        raise ValueError(f"Cannot find a component type suitable for {dt}")
    return cast(int, val)


class GltfPrimitive:
    def __init__(
        self,
        points: npt.NDArray[np.float32],
        triangles: npt.NDArray[np.uint8 | np.uint16 | np.uint32] | None = None,
        normals: npt.NDArray[np.float32] | None = None,
        uvs: npt.NDArray[np.float32] | None = None,
        batchids: npt.NDArray[np.uint32] | None = None,
        additional_attributes: list[GltfAttribute] | None = None,
        texture_uri: str | None = None,
        material: pygltflib.Material | None = None,
    ) -> None:
        """
        A data structure storing all information to create a glTF mesh's primitive.

        :param points: array of vertex positions, must have a (n, 3) shape.
        :param triangles: array of triangle indices, must have a (n, 3) shape.
        :param normals: array of vertex normals, must have a (n, 3) shape.
        :param uvs: array of texture coordinates, must have a (n, 2) shape.
        :param batchids: array of batch table IDs, must have a (n) shape.
        :param additional_attributes: additional attributes to add to the primitive.
        :param texture_uri: the URI of the texture image if the primitive is textured.
        :param material: a glTF material. If not set, a default material is created.
        """
        self.points: GltfAttribute = GltfAttribute(
            "POSITION", pygltflib.VEC3, pygltflib.FLOAT, points
        )
        self.triangles: GltfAttribute | None = (
            GltfAttribute(
                "INDICE",
                pygltflib.SCALAR,
                get_component_type_from_dtype(triangles.dtype),
                triangles,
            )
            if triangles is not None
            else None
        )
        self.normals: GltfAttribute | None = (
            GltfAttribute("NORMAL", pygltflib.VEC3, pygltflib.FLOAT, normals)
            if normals is not None
            else None
        )
        self.uvs: GltfAttribute | None = (
            GltfAttribute("TEXCOORD_0", pygltflib.VEC2, pygltflib.FLOAT, uvs)
            if uvs is not None
            else None
        )
        self.batchids: GltfAttribute | None = (
            GltfAttribute(
                "_BATCHID", pygltflib.SCALAR, pygltflib.UNSIGNED_INT, batchids
            )
            if batchids is not None
            else None
        )
        self.additional_attributes: list[GltfAttribute] = (
            additional_attributes if additional_attributes is not None else []
        )
        self.texture_uri: str | None = texture_uri
        self.material: pygltflib.Material | None = material


def gltf_component_from_primitive(
    primitive: GltfPrimitive,
    byte_offset: int = 0,
    attribute_counter: int = 0,
) -> tuple[
    pygltflib.Primitive, list[pygltflib.Accessor], list[pygltflib.BufferView], bytes
]:
    """Build the GlTF structure that corresponds to a triangulated mesh.

    The mesh is represented as numpy arrays as follows:

    - a 3D-point array;
    - a triangle array, where points are identified by their positional ID in the 3D-point
      array.

    See https://gitlab.com/dodgyville/pygltflib/ (section "Create a mesh, convert to bytes,
    convert back to mesh").

    """
    gltf_primitive = pygltflib.Primitive(
        attributes=pygltflib.Attributes(),
    )
    gltf_binary_blob = b""
    gltf_accessors = []
    gltf_buffer_views = []

    if primitive.triangles is not None:
        indice_blob, indice_accessor, indice_buffer_view = prepare_gltf_component(
            attribute_counter,
            primitive.triangles.array,
            byte_offset,
            primitive.triangles.array.size,
            primitive.triangles.accessor_type,
            primitive.triangles.component_type,
            pygltflib.ELEMENT_ARRAY_BUFFER,
        )
        gltf_binary_blob += indice_blob
        gltf_accessors.append(indice_accessor)
        gltf_buffer_views.append(indice_buffer_view)
        gltf_primitive.indices = attribute_counter
        attribute_counter += 1

    attributes_array: list[GltfAttribute | None] = [
        primitive.points,
        primitive.normals,
        primitive.uvs,
        primitive.batchids,
    ]
    attributes_array.extend(primitive.additional_attributes)

    for attribute in attributes_array:
        if attribute is None:
            continue
        (array_blob, accessor, buffer_view,) = prepare_gltf_component(
            attribute_counter,
            attribute.array,
            byte_offset + len(gltf_binary_blob),
            len(attribute.array),
            attribute.accessor_type,
            attribute.component_type,
        )
        gltf_binary_blob += array_blob
        gltf_accessors.append(accessor)
        gltf_buffer_views.append(buffer_view)
        setattr(gltf_primitive.attributes, attribute.name, attribute_counter)
        attribute_counter += 1
    gltf_accessors[int(primitive.triangles is not None)].min = np.min(
        primitive.points.array, axis=0
    ).tolist()
    gltf_accessors[int(primitive.triangles is not None)].max = np.max(
        primitive.points.array, axis=0
    ).tolist()

    return gltf_primitive, gltf_accessors, gltf_buffer_views, gltf_binary_blob


def prepare_gltf_component(
    array_idx: int,
    array: npt.NDArray[np.uint8 | np.uint16 | np.uint32 | np.float32],
    byte_offset: int,
    count: int,
    accessor_type: str = pygltflib.VEC3,
    component_type: int = pygltflib.FLOAT,
    buffer_view_target: int = pygltflib.ARRAY_BUFFER,
) -> tuple[bytes, pygltflib.Accessor, pygltflib.BufferView]:
    array_blob = array.flatten().tobytes()
    BUFFER_INDEX = 0  # Everything is stored in the same buffer for sake of simplicity
    accessor = pygltflib.Accessor(
        bufferView=array_idx,
        componentType=component_type,
        count=count,
        type=accessor_type,
    )
    buffer_view = pygltflib.BufferView(
        buffer=BUFFER_INDEX,
        byteOffset=byte_offset,
        byteLength=len(array_blob),
        target=buffer_view_target,
    )
    return array_blob, accessor, buffer_view



import math
from pathlib import Path
from typing import Generator, List, Optional, Tuple

import laspy
import numpy as np
import numpy.typing as npt
from pyproj import Transformer

from py3dtiles.typing import MetadataReaderType, OffsetScaleType, PortionItemType


def get_metadata(path: Path, fraction: int = 100) -> MetadataReaderType:
    pointcloud_file_portions = []

    filename = str(path)
    with laspy.open(filename) as f:
        point_count = f.header.point_count * fraction // 100

        _1M = min(point_count, 1_000_000)
        steps = math.ceil(point_count / _1M)
        portions: List[PortionItemType] = [
            (i * _1M, min(point_count, (i + 1) * _1M)) for i in range(steps)
        ]
        for p in portions:
            pointcloud_file_portions += [(filename, p)]

        crs_in = f.header.parse_crs()

    return {
        "portions": pointcloud_file_portions,
        "aabb": np.array([f.header.mins, f.header.maxs]),
        "crs_in": crs_in,
        "point_count": point_count,
        "avg_min": np.array(f.header.mins),
    }


def run(
    filename: str,
    offset_scale: OffsetScaleType,
    portion: PortionItemType,
    transformer: Optional[Transformer],
    color_scale: Optional[float],
    write_intensity: bool,
) -> Generator[
    Tuple[
        npt.NDArray[np.float32],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
    ],
    None,
    None,
]:
    """
    Reads points from a las file
    """
    with laspy.open(filename) as f:

        point_count = portion[1] - portion[0]

        step = min(point_count, max(point_count // 10, 100_000))

        indices = list(range(math.ceil(point_count / step)))

        for index in indices:
            start_offset = portion[0] + index * step
            num = min(step, portion[1] - start_offset)

            # read scaled values and apply offset
            f.seek(start_offset)
            points = next(f.chunk_iterator(num))

            x, y, z = points.x, points.y, points.z
            if transformer:
                x, y, z = transformer.transform(x, y, z)

            x = (x + offset_scale[0][0]) * offset_scale[1][0]
            y = (y + offset_scale[0][1]) * offset_scale[1][1]
            z = (z + offset_scale[0][2]) * offset_scale[1][2]

            coords = np.vstack((x, y, z)).transpose()

            if offset_scale[2] is not None:
                # Apply transformation matrix (because the tile's transform will contain
                # the inverse of this matrix)
                coords = np.dot(coords, offset_scale[2])

            coords = np.ascontiguousarray(coords.astype(np.float32))

            # Read colors
            if "red" in f.header.point_format.dimension_names:
                red = points["red"]
                green = points["green"]
                blue = points["blue"]
            else:
                red = points["intensity"]
                green = points["intensity"]
                blue = points["intensity"]

            colors = np.vstack((red, green, blue)).transpose()

            if color_scale is not None:
                colors = np.clip(colors * color_scale, 0, 65535)

            # NOTE las spec says rgb is 16bits by components
            # pnts are 8 bits (by default) by component, hence we divide by 256
            colors = (colors / 256).astype(np.uint8)

            if "classification" in f.header.point_format.dimension_names:
                classification = np.array(
                    points["classification"], dtype=np.uint8
                ).reshape(-1, 1)
            else:
                classification = np.zeros((len(points.x), 1), dtype=np.uint8)

            if "intensity" in f.header.point_format.dimension_names:
                intensity = np.array(points["intensity"] / 256, dtype=np.uint8).reshape(
                    -1, 1
                )
            else:
                intensity = np.zeros((len(points.x), 1), dtype=np.uint8)

            yield coords, colors, classification, intensity



from __future__ import annotations

import pickle
from pathlib import Path
from typing import TYPE_CHECKING, Generator

import lz4.frame as gzip
import numpy as np
import numpy.typing as npt

import py3dtiles
from py3dtiles.tileset.content import Pnts
from py3dtiles.utils import node_name_to_path

if TYPE_CHECKING:
    from py3dtiles.tilers.point.node import DummyNode, Node


def points_to_pnts_file(
    name: bytes,
    points: npt.NDArray[np.uint8],
    out_folder: Path,
    include_rgb: bool,
    include_classification: bool,
    include_intensity: bool,
) -> tuple[int, Path]:
    """
    Write a pnts file from an uint8 data array containing:
     - points as SemanticPoint.POSITION
     - if include_rgb, rgb as SemanticPoint.RGB
     - if include_classification, classification as a single np.uint8 value
     - if include_intensity, intensity as a single np.uint8 value
    """
    pnts = Pnts.from_points(
        points, include_rgb, include_classification, include_intensity
    )

    node_path = node_name_to_path(out_folder, name, ".pnts")

    if node_path.exists():
        raise FileExistsError(f"{node_path} already written")

    pnts.save_as(node_path)

    return pnts.body.feature_table.nb_points(), node_path


def node_to_pnts(
    name: bytes,
    node: Node | DummyNode,
    out_folder: Path,
    include_rgb: bool,
    include_classification: bool,
    include_intensity: bool,
) -> int:
    points = py3dtiles.tilers.point.node.Node.get_points(
        node, include_rgb, include_classification, include_intensity
    )
    point_nb, _ = points_to_pnts_file(
        name, points, out_folder, include_rgb, include_classification, include_intensity
    )
    return point_nb


def run(
    data: bytes,
    folder: Path,
    write_rgb: bool,
    write_classification: bool,
    write_intensity: bool,
) -> Generator[int, None, None]:
    # we can safely write the .pnts file
    if len(data) > 0:
        root = pickle.loads(gzip.decompress(data))
        total = 0
        for name in root:
            node = py3dtiles.tilers.point.node.DummyNode(pickle.loads(root[name]))
            total += node_to_pnts(
                name, node, folder, write_rgb, write_classification, write_intensity
            )
        yield total



import csv
import math
from pathlib import Path
from typing import Generator, List, Optional, Tuple

import numpy as np
import numpy.typing as npt
from pyproj import Transformer

from py3dtiles.typing import MetadataReaderType, OffsetScaleType, PortionItemType


def get_metadata(path: Path, fraction: int = 100) -> MetadataReaderType:
    aabb = None
    count = 0
    seek_values = []

    with path.open() as f:
        file_sample = f.read(
            2048
        )  # For performance reasons we just snif the first part
        dialect = csv.Sniffer().sniff(file_sample)

        f.seek(0)
        if csv.Sniffer().has_header(file_sample):
            f.readline()

        while True:
            batch = 10_000
            points = np.zeros((batch, 3))

            offset = f.tell()
            for i in range(batch):
                line = f.readline()
                if not line:
                    points = np.resize(points, (i, 3))
                    break
                points[i] = [float(s) for s in line.split(dialect.delimiter)][:3]

            if points.shape[0] == 0:
                break

            if not count % 1_000_000:
                seek_values += [offset]

            count += points.shape[0]
            batch_aabb = np.array([np.min(points, axis=0), np.max(points, axis=0)])

            # Update aabb
            if aabb is None:
                aabb = batch_aabb
            else:
                aabb[0] = np.minimum(aabb[0], batch_aabb[0])
                aabb[1] = np.maximum(aabb[1], batch_aabb[1])

        # We need an exact point count
        point_count = count * fraction // 100

        _1M = min(count, 1_000_000)
        steps = math.ceil(count / _1M)
        if steps != len(seek_values):
            raise ValueError(
                "the size of seek_values should be equal to steps,"
                f"currently steps = {steps} and len(seek_values) = {len(seek_values)}"
            )
        portions: List[PortionItemType] = [
            (i * _1M, min(count, (i + 1) * _1M), seek_values[i]) for i in range(steps)
        ]

        pointcloud_file_portions = [(str(path), p) for p in portions]

    if aabb is None:
        raise ValueError(f"There is no point in the file {path}")

    return {
        "portions": pointcloud_file_portions,
        "aabb": aabb,
        "crs_in": None,
        "point_count": point_count,
        "avg_min": aabb[0],
    }


def run(
    filename: str,
    offset_scale: OffsetScaleType,
    portion: PortionItemType,
    transformer: Optional[Transformer],
    color_scale: Optional[float],
    write_intensity: bool,
) -> Generator[
    Tuple[
        npt.NDArray[np.float32],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
    ],
    None,
    None,
]:
    """
    Reads points from a .xyz or .csv file

    Consider XYZIRGB format following FME documentation(*). We do the
    following hypothesis and enhancements:

    - A header line defining columns in CSV style may be present, but will be ignored.
    - The separator separating the columns is automagically guessed by the
      reader. This is generally fail safe. It will not harm to use commonly
      accepted separators like space, tab, colon, semi-colon.
    - The order of columns is fixed. The reader does the following assumptions:
      - 3 columns mean XYZ
      - 4 columns mean XYZI
      - 6 columns mean XYZRGB
      - 7 columns mean XYZIRGB
      - 8 columns mean XYZIRGB followed by classification data. Classification data must be integers only.
      - all columns after the 8th column will be ignored.

    NOTE: we assume RGB are 8 bits components.

    (*) See: https://docs.safe.com/fme/html/FME_Desktop_Documentation/FME_ReadersWriters/pointcloudxyz/pointcloudxyz.htm
    """
    with open(filename) as f:

        dialect = csv.Sniffer().sniff(f.read(2048))
        f.seek(0)
        f.readline()  # skip first line in case there is a header we promised to ignore
        feature_nb = len(f.readline().split(dialect.delimiter))
        if feature_nb < 8:
            feature_nb = 7  # we pad to 7 columns with 0
        if feature_nb > 8:
            feature_nb = 8  # We ignore other data as downstream only 1 value for classification data is supported.
            # Once downstream supports multiple classification values this reader will as well
            # when this line is removed.

        point_count = portion[1] - portion[0]

        step = min(point_count, max((point_count) // 10, 100000))

        f.seek(portion[2])

        for _ in range(0, point_count, step):
            points = np.zeros((step, feature_nb), dtype=np.float32)

            for j in range(step):
                line = f.readline()
                if not line:
                    points = np.resize(points, (j, feature_nb))
                    break
                line_features: List[Optional[float]] = [
                    float(s) for s in line.split(dialect.delimiter)[:feature_nb]
                ]
                if len(line_features) == 3:
                    line_features += [None] * 4  # Insert intensity and RGB
                elif len(line_features) == 4:
                    line_features += [None] * 3  # Insert RGB
                elif len(line_features) == 6:
                    line_features.insert(3, None)  # Insert intensity
                points[j] = line_features

            x, y, z = (points[:, c] for c in [0, 1, 2])

            if transformer:
                x, y, z = transformer.transform(x, y, z)

            x = (x + offset_scale[0][0]) * offset_scale[1][0]
            y = (y + offset_scale[0][1]) * offset_scale[1][1]
            z = (z + offset_scale[0][2]) * offset_scale[1][2]

            coords = np.vstack((x, y, z)).transpose()

            if offset_scale[2] is not None:
                # Apply transformation matrix (because the tile's transform will contain
                # the inverse of this matrix)
                coords = np.dot(coords, offset_scale[2])

            coords = np.ascontiguousarray(coords.astype(np.float32))

            # Read colors: 3 last columns when excluding classification data
            if color_scale is None:
                colors = points[:, 4:7].astype(np.uint8)
            else:
                colors = np.clip(points[:, 4:7] * color_scale, 0, 255).astype(np.uint8)

            if feature_nb > 7:  # we have classification data
                classification = np.array(points[:, 7:], dtype=np.uint8).reshape(-1, 1)
            else:
                classification = np.zeros((points.shape[0], 1), dtype=np.uint8)

            if feature_nb in (4, 7, 8) and write_intensity:
                intensity = np.array(points[:, 3], dtype=np.uint8).reshape(-1, 1)
            else:
                intensity = np.zeros((points.shape[0], 1), dtype=np.uint8)

            yield coords, colors, classification, intensity



import shutil
from pathlib import Path

import numpy as np
from pytest import raises

from py3dtiles.utils import make_aabb_valid, mkdir_or_raise


def test_make_aabb_valid() -> None:
    aabb = np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
    make_aabb_valid(aabb)
    np.testing.assert_array_equal(aabb, np.array([[0, 0, 0], [1.0, 1.0, 1.0]]))
    aabb = np.array([[0, 0, 0], [0, 0, 0]], dtype=np.float64)
    make_aabb_valid(aabb)
    np.testing.assert_array_equal(
        aabb, np.array([[0, 0, 0], [0.00001, 0.00001, 0.00001]])
    )
    aabb = np.array([[9, 10, 11], [9, 10, 11]], dtype=np.float64)
    make_aabb_valid(aabb)
    np.testing.assert_array_equal(
        aabb, np.array([[9, 10, 11], [9.00001, 10.00001, 11.00001]])
    )
    aabb = np.array([[9, 12, 14], [10, 13, 15]], dtype=np.float64)
    make_aabb_valid(aabb)
    np.testing.assert_array_equal(aabb, np.array([[9, 12, 14], [10, 13, 15]]))
    aabb = np.array([[9, 12, 14], [9, 13, 15]], dtype=np.float64)
    make_aabb_valid(aabb)
    np.testing.assert_array_equal(aabb, np.array([[9, 12, 14], [9.00001, 13, 15]]))


def test_mkdir_or_raise(tmp_dir: Path) -> None:
    mkdir_or_raise(tmp_dir)
    (tmp_dir / "foo").touch()

    with raises(
        FileExistsError, match=f"Folder '{tmp_dir}' already exists and is not empty."
    ):
        mkdir_or_raise(tmp_dir)

    mkdir_or_raise(tmp_dir, overwrite=True)
    assert len(list(tmp_dir.iterdir())) == 0

    shutil.rmtree(tmp_dir)

    mkdir_or_raise(tmp_dir)
    # second time, should still succeed because it is empty
    mkdir_or_raise(tmp_dir)

    shutil.rmtree(tmp_dir)
    # it is now a file
    tmp_dir.touch()

    with raises(
        FileExistsError,
        match=f"'{tmp_dir}' already exists and is not a directory. Not deleting it.",
    ):
        mkdir_or_raise(tmp_dir)
    tmp_dir.unlink()



from enum import Enum


class ManagerMessage(Enum):
    """
    Message types sent by the convert function.

    If you create a new tiler, this category of messages is sent by Tiler.
    You cannot inherit this class since it is an enum. Create a new one without these 2 message types.
    """

    STOP = b"stop"
    SHUTDOWN = b"shutdown"


class WorkerMessageType(Enum):
    """
    Message types sent by the broker.

    If you create a new tiler, this category of messages is sent by WorkerTiler.
    You cannot inherit this class. Create a new one without these 3 message types.
    """

    REGISTER = b"register"
    IDLE = b"idle"
    HALTED = b"halted"
    ERROR = b"error"



from typing import Dict, List, Tuple

from py3dtiles.typing import PortionsType


class PointState:
    def __init__(
        self, pointcloud_file_portions: PortionsType, max_reading_jobs: int
    ) -> None:
        self.processed_points = 0
        self.max_point_in_progress = 60_000_000
        self.points_in_progress = 0
        self.points_in_pnts = 0

        # pointcloud_file_portions is a list of tuple (filename, (start offset, end offset))
        self.point_cloud_file_parts = pointcloud_file_portions
        self.initial_portion_count = len(pointcloud_file_portions)
        self.max_reading_jobs = max_reading_jobs
        self.number_of_reading_jobs = 0
        self.number_of_writing_jobs = 0

        # node_to_process is a dictionary of tasks,
        # each entry is a tile identified by its name (a string of numbers)
        # so for each entry, it is a list of tasks
        # a task is a tuple (list of points, point_count)
        # points is a dictionary {xyz: list of coordinates, color: the associated color}
        self.node_to_process: Dict[bytes, Tuple[List[bytes], int]] = {}
        # when a node is sent to a process, the item moves to processing_nodes
        # the structure is different. The key remains the node name. But the value is : (len(tasks), point_count, now)
        # these values is for logging
        self.processing_nodes: Dict[bytes, Tuple[int, int, float]] = {}
        # when processing is finished, move the tile name in processed_nodes
        # since the content is at this stage, stored in the node_store,
        # just keep the name of the node.
        # This list will be filled until the writing could be started.
        self.waiting_writing_nodes: List[bytes] = []
        # when the node is writing, its name is moved from waiting_writing_nodes to pnts_to_writing
        # the data to write are stored in a node object.
        self.pnts_to_writing: List[bytes] = []

    def is_reading_finish(self) -> bool:
        return not self.point_cloud_file_parts and self.number_of_reading_jobs == 0

    def add_tasks_to_process(
        self, node_name: bytes, data: bytes, point_count: int
    ) -> None:
        if point_count <= 0:
            raise ValueError(
                "point_count should be strictly positive, currently", point_count
            )

        if node_name not in self.node_to_process:
            self.node_to_process[node_name] = ([data], point_count)
        else:
            tasks, count = self.node_to_process[node_name]
            tasks.append(data)
            self.node_to_process[node_name] = (tasks, count + point_count)

    def can_add_reading_jobs(self) -> bool:
        return bool(
            self.point_cloud_file_parts
            and self.points_in_progress < self.max_point_in_progress
            and self.number_of_reading_jobs < self.max_reading_jobs
        )

    def print_debug(self) -> None:
        print("{:^16}|{:^8}|{:^8}|{:^8}".format("Step", "Input", "Active", "Inactive"))
        print(
            "{:^16}|{:^8}|{:^8}|{:^8}".format(
                "Reader",
                len(self.point_cloud_file_parts),
                self.number_of_reading_jobs,
                "",
            )
        )
        print(
            "{:^16}|{:^8}|{:^8}|{:^8}".format(
                "Node process",
                len(self.node_to_process),
                len(self.processing_nodes),
                len(self.waiting_writing_nodes),
            )
        )
        print(
            "{:^16}|{:^8}|{:^8}|{:^8}".format(
                "Pnts writer",
                len(self.pnts_to_writing),
                self.number_of_writing_jobs,
                "",
            )
        )



"""
This module contains all the exceptions that py3dtiles generates. They all derive from :py:class:`Py3dtilesException`.
"""


class Py3dtilesException(Exception):
    """
    All exceptions thrown by py3dtiles code derives this class.

    Client code that wishes to catch all py3dtiles exception can use `except Py3dtilesException`.
    """


class TilerException(Py3dtilesException):
    """
    This exception will be thrown when there is an issue during a tiling task.
    """


class WorkerException(TilerException):
    """
    This exception will be thrown by the conversion code if one exception occurs inside a worker.
    """


class SrsInMissingException(Py3dtilesException):
    """
    This exception will be thrown when an input srs is required but not provided.
    """


class SrsInMixinException(Py3dtilesException):
    """
    This exception will be thrown when among all input files, there is a mix of input srs.
    """


class Invalid3dtilesError(Py3dtilesException):
    """
    This exception will be thrown if the 3d tile specification isn't respected.
    """


class InvalidPntsError(Invalid3dtilesError):
    """
    This exception will be thrown if the point cloud format isn't respected.
    """


class InvalidB3dmError(Invalid3dtilesError):
    """
    This exception will be thrown if the batched 3D model format isn't respected.
    """


class InvalidTilesetError(Invalid3dtilesError):
    """
    This exception will be thrown if the tileset format isn't respected.
    """


class BoundingVolumeMissingException(InvalidTilesetError):
    """
    This exception will be thrown when a bounding volume is needed but not present.
    """



from enum import Enum


class PointManagerMessage(Enum):
    READ_FILE = b"read_file"
    WRITE_PNTS = b"write_pnts"
    PROCESS_JOBS = b"process_job"


class PointWorkerMessageType(Enum):
    READ = b"read"
    PROCESSED = b"processed"
    PNTS_WRITTEN = b"pnts_written"
    NEW_TASK = b"new_task"



from __future__ import annotations

import copy
import json
import pickle
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Iterator, TypedDict

import numpy as np
import numpy.typing as npt

from py3dtiles.exceptions import TilerException
from py3dtiles.tilers.point.pnts import MIN_POINT_SIZE
from py3dtiles.tilers.point.pnts.pnts_writer import points_to_pnts_file
from py3dtiles.tileset.bounding_volume_box import BoundingVolumeBox
from py3dtiles.tileset.content import read_binary_tile_content
from py3dtiles.tileset.content.pnts_feature_table import SemanticPoint
from py3dtiles.tileset.tile import Tile
from py3dtiles.tileset.tileset import TileSet
from py3dtiles.utils import (
    SubdivisionType,
    aabb_size_to_subdivision_type,
    node_from_name,
    node_name_to_path,
)

from .distance import xyz_to_child_index
from .points_grid import Grid

if TYPE_CHECKING:
    from typing_extensions import NotRequired

    from .node_catalog import NodeCatalog


def node_to_tileset(
    args: tuple[Node, Path, npt.NDArray[np.float32], Node | None, int]
) -> Tile | None:
    return args[0].to_tileset(args[1], args[2], args[3], args[4], None)


class _DummyNodeDictType(TypedDict):
    children: NotRequired[list[bytes]]
    grid: NotRequired[Grid]
    points: NotRequired[
        list[
            tuple[
                npt.NDArray[np.float32],
                npt.NDArray[np.uint8],
                npt.NDArray[np.uint8],
                npt.NDArray[np.uint8],
            ]
        ]
    ]


class DummyNode:
    def __init__(self, _bytes: _DummyNodeDictType) -> None:
        if "children" in _bytes:
            self.children: list[bytes] | None = _bytes["children"]
            self.grid = _bytes["grid"]
        else:
            self.children = None
            self.points = _bytes["points"]


class Node:
    """docstring for Node"""

    __slots__ = (
        "name",
        "aabb",
        "aabb_size",
        "inv_aabb_size",
        "aabb_center",
        "spacing",
        "pending_xyz",
        "pending_rgb",
        "pending_classification",
        "pending_intensity",
        "children",
        "grid",
        "points",
        "dirty",
    )

    def __init__(
        self, name: bytes, aabb: npt.NDArray[np.float64 | np.float32], spacing: float
    ) -> None:
        super().__init__()
        self.name = name
        self.aabb = aabb.astype(
            np.float32
        )  # TODO remove astype once the whole typing is done (and once data type issues on numpy arrays are fixed).
        self.aabb_size = np.maximum(self.aabb[1] - self.aabb[0], MIN_POINT_SIZE)
        self.inv_aabb_size = 1.0 / self.aabb_size
        self.aabb_center = (self.aabb[0] + self.aabb[1]) * 0.5
        self.spacing = spacing
        self.pending_xyz: list[npt.NDArray[np.float32]] = []
        self.pending_rgb: list[npt.NDArray[np.uint8]] = []
        self.pending_classification: list[npt.NDArray[np.uint8]] = []
        self.pending_intensity: list[npt.NDArray[np.uint8]] = []
        self.children: list[bytes] | None = None
        self.grid = Grid(self)
        self.points: list[
            tuple[
                npt.NDArray[np.float32],
                npt.NDArray[np.uint8],
                npt.NDArray[np.uint8],
                npt.NDArray[np.uint8],
            ]
        ] = []
        self.dirty = False

    def save_to_bytes(self) -> bytes:
        sub_pickle: dict[str, Any] = {}
        if self.children is not None:
            sub_pickle["children"] = self.children
            sub_pickle["grid"] = self.grid
        else:
            sub_pickle["points"] = self.points

        return pickle.dumps(sub_pickle)

    def load_from_bytes(self, byt: bytes) -> None:
        sub_pickle = pickle.loads(byt)
        if "children" in sub_pickle:
            self.children = sub_pickle["children"]
            self.grid = sub_pickle["grid"]
        else:
            self.points = sub_pickle["points"]

    def insert(
        self,
        scale: float,
        xyz: npt.NDArray[np.float32],
        rgb: npt.NDArray[np.uint8],
        classification: npt.NDArray[np.uint8],
        intensity: npt.NDArray[np.uint8],
        make_empty_node: bool = False,
    ) -> None:
        if make_empty_node:
            self.children = []
            self.pending_xyz += [xyz]
            self.pending_rgb += [rgb]
            self.pending_classification += [classification]
            self.pending_intensity += [intensity]
            return

        # fastpath
        if self.children is None:
            self.points.append((xyz, rgb, classification, intensity))
            count = sum([xyz.shape[0] for xyz, _, _, _ in self.points])
            # stop subdividing if spacing is 1mm
            if count >= 20000 and self.spacing > 0.001 * scale:
                self._split(scale)
            self.dirty = True

            return

        # grid based insertion
        (
            remainder_xyz,
            remainder_rgb,
            remainder_classification,
            remainder_intensity,
            needs_balance,
        ) = self.grid.insert(
            self.aabb[0], self.inv_aabb_size, xyz, rgb, classification, intensity
        )

        if needs_balance:
            self.grid.balance(self.aabb_size, self.aabb[0], self.inv_aabb_size)
            self.dirty = True

        self.dirty = self.dirty or (len(remainder_xyz) != len(xyz))

        if len(remainder_xyz) > 0:
            self.pending_xyz += [remainder_xyz]
            self.pending_rgb += [remainder_rgb]
            self.pending_classification += [remainder_classification]
            self.pending_intensity += [remainder_intensity]

    def needs_balance(self) -> bool:
        if self.children is not None:
            return self.grid.needs_balance()
        return False

    def flush_pending_points(self, catalog: NodeCatalog, scale: float) -> None:
        for name, xyz, rgb, classification, intensity in self._get_pending_points():
            catalog.get_node(name).insert(scale, xyz, rgb, classification, intensity)
        self.pending_xyz = []
        self.pending_rgb = []
        self.pending_classification = []
        self.pending_intensity = []

    def dump_pending_points(self) -> list[tuple[bytes, bytes, int]]:
        result = [
            (
                name,
                pickle.dumps(
                    {
                        "xyz": xyz,
                        "rgb": rgb,
                        "classification": classification,
                        "intensity": intensity,
                    }
                ),
                len(xyz),
            )
            for name, xyz, rgb, classification, intensity in self._get_pending_points()
        ]

        self.pending_xyz = []
        self.pending_rgb = []
        self.pending_classification = []
        self.pending_intensity = []
        return result

    def get_pending_points_count(self) -> int:
        return sum([xyz.shape[0] for xyz in self.pending_xyz])

    def _get_pending_points(
        self,
    ) -> Iterator[
        tuple[
            bytes,
            npt.NDArray[np.float32],
            npt.NDArray[np.uint8],
            npt.NDArray[np.uint8],
            npt.NDArray[np.uint8],
        ]
    ]:
        if not self.pending_xyz:
            return

        pending_xyz_arr = np.concatenate(self.pending_xyz)
        pending_rgb_arr = np.concatenate(self.pending_rgb)
        pending_classification_arr = np.concatenate(self.pending_classification)
        pending_intensity_arr = np.concatenate(self.pending_intensity)
        t = aabb_size_to_subdivision_type(self.aabb_size)
        if t == SubdivisionType.QUADTREE:
            indices = xyz_to_child_index(
                pending_xyz_arr,
                np.array(
                    [self.aabb_center[0], self.aabb_center[1], self.aabb[1][2]],
                    dtype=np.float32,
                ),
            )
        else:
            indices = xyz_to_child_index(pending_xyz_arr, self.aabb_center)

        # unique children list
        childs = np.unique(indices)

        # make sure all children nodes exist
        for child in childs:
            name = "{}{}".format(self.name.decode("ascii"), child).encode("ascii")
            # create missing nodes, only for remembering they exist.
            # We don't want to serialize them
            # probably not needed...
            if self.children is not None and name not in self.children:
                self.children += [name]
                self.dirty = True
                # print('Added node {}'.format(name))

            mask = np.where(indices - child == 0)
            xyz = pending_xyz_arr[mask]
            if len(xyz) > 0:
                yield name, xyz, pending_rgb_arr[mask], pending_classification_arr[
                    mask
                ], pending_intensity_arr[mask]

    def _split(self, scale: float) -> None:
        self.children = []
        for xyz, rgb, classification, intensity in self.points:
            self.insert(scale, xyz, rgb, classification, intensity)
        self.points = []

    def get_point_count(
        self, node_catalog: NodeCatalog, max_depth: int, depth: int = 0
    ) -> int:
        if self.children is None:
            return sum([xyz.shape[0] for xyz, _, _, _ in self.points])
        else:
            count = self.grid.get_point_count()
            if depth < max_depth:
                for n in self.children:
                    count += node_catalog.get_node(n).get_point_count(
                        node_catalog, max_depth, depth + 1
                    )
            return count

    @staticmethod
    def get_points(
        data: Node | DummyNode,
        include_rgb: bool,
        include_classification: bool,
        include_intensity: bool,
    ) -> npt.NDArray[np.uint8]:  # todo remove staticmethod
        if data.children is None:
            points = data.points
            xyz = (
                np.concatenate(tuple([xyz for xyz, _, _, _ in points]))
                .view(np.uint8)
                .ravel()
            )

            if include_rgb:
                rgb = np.concatenate(tuple([rgb for _, rgb, _, _ in points])).ravel()
            else:
                rgb = np.array([], dtype=np.uint8)

            if include_classification:
                classification = np.concatenate(
                    tuple([classification for _, _, classification, _ in points])
                ).ravel()
            else:
                classification = np.array([], dtype=np.uint8)

            if include_intensity:
                intensity = np.concatenate(
                    tuple([intensity for _, _, _, intensity in points])
                ).ravel()
            else:
                intensity = np.array([], dtype=np.uint8)

            return np.concatenate((xyz, rgb, classification, intensity))
        else:
            return data.grid.get_points(
                include_rgb, include_classification, include_intensity
            )

    def get_child_names(self) -> Generator[bytes, None, None]:
        for number_child in range(8):
            yield f"{self.name.decode('ascii')}{number_child}".encode("ascii")

    def to_tileset(
        self,
        folder: Path,
        scale: npt.NDArray[np.float32],
        parent_node: Node | None = None,
        depth: int = 0,
        pool_executor: ProcessPoolExecutor | None = None,
    ) -> Tile | None:
        # create child tileset parts
        # if their size is below of 100 points, they will be merged in this node.
        children_tileset_parts: list[Tile] = []
        parameter_to_compute: list[
            tuple[Node, Path, npt.NDArray[np.float32], Node, int]
        ] = []
        for child_name in self.get_child_names():
            child_node = node_from_name(child_name, self.aabb, self.spacing)
            child_pnts_path = node_name_to_path(folder, child_name, ".pnts")

            if child_pnts_path.exists():
                # multi thread is only allowed on nodes where there are no prune
                # a simple rule is: only is there is not a parent node
                if pool_executor and parent_node is None:
                    parameter_to_compute.append(
                        (child_node, folder, scale, self, depth + 1)
                    )
                else:
                    children_tileset_part = child_node.to_tileset(
                        folder, scale, self, depth + 1
                    )
                    if (
                        children_tileset_part is not None
                    ):  # return None if the child has been merged
                        children_tileset_parts.append(children_tileset_part)

        if pool_executor and parent_node is None:
            children_tileset_parts = [
                t
                for t in pool_executor.map(node_to_tileset, parameter_to_compute)
                if t is not None
            ]

        pnts_path = node_name_to_path(folder, self.name, ".pnts")
        tile_content = read_binary_tile_content(pnts_path)
        fth = tile_content.body.feature_table.header
        xyz = tile_content.body.feature_table.body.position

        # check if this node should be merged in the parent.
        prune = False  # prune only if the node is a leaf

        # If this child is small enough, merge in the current tile
        if parent_node is not None and depth > 1 and fth.points_length < 100:
            parent_pnts_path = node_name_to_path(folder, parent_node.name, ".pnts")
            parent_tile = read_binary_tile_content(parent_pnts_path)
            parent_fth = parent_tile.body.feature_table.header

            parent_xyz = parent_tile.body.feature_table.body.position

            if (
                parent_fth.colors != SemanticPoint.NONE
                and parent_tile.body.feature_table.body.color is not None
            ):
                parent_rgb = parent_tile.body.feature_table.body.color
            else:
                parent_rgb = np.array([], dtype=np.uint8)

            if "Classification" in parent_tile.body.batch_table.header.data:
                parent_classification = (
                    parent_tile.body.batch_table.get_binary_property("Classification")
                )
            else:
                parent_classification = np.array([], dtype=np.uint8)

            if "Intensity" in parent_tile.body.batch_table.header.data:
                parent_intensity = parent_tile.body.batch_table.get_binary_property(
                    "Intensity"
                )
            else:
                parent_intensity = np.array([], dtype=np.uint8)

            parent_xyz_float = parent_xyz.reshape((parent_fth.points_length, 3))
            # update aabb based on real values
            parent_bounding_volume = BoundingVolumeBox.from_points(parent_xyz_float)

            parent_xyz = np.concatenate((parent_xyz, xyz))

            if fth.colors != SemanticPoint.NONE:
                if tile_content.body.feature_table.body.color is None:
                    raise TilerException(
                        "If the parent has color data, the children must also have color data."
                    )
                parent_rgb = np.concatenate(
                    (parent_rgb, tile_content.body.feature_table.body.color)
                )

            if "Classification" in tile_content.body.batch_table.header.data:
                parent_classification = np.concatenate(
                    (
                        parent_classification,
                        tile_content.body.batch_table.get_binary_property(
                            "Classification"
                        ),
                    )
                )

            if "Intensity" in tile_content.body.batch_table.header.data:
                parent_intensity = np.concatenate(
                    (
                        parent_intensity,
                        tile_content.body.batch_table.get_binary_property("Intensity"),
                    )
                )

            # update aabb
            xyz_float = xyz.view(np.float32).reshape((fth.points_length, 3))
            new_bounding_volume_box = BoundingVolumeBox.from_points(xyz_float)

            parent_bounding_volume.add(new_bounding_volume_box)

            parent_pnts_path.unlink()
            points_to_pnts_file(
                parent_node.name,
                np.concatenate(
                    (
                        parent_xyz.view(np.uint8),
                        parent_rgb,
                        parent_classification,
                        parent_intensity,
                    )
                ),
                folder,
                len(parent_rgb) != 0,
                len(parent_classification) != 0,
                len(parent_intensity) != 0,
            )
            pnts_path.unlink()
            prune = True

        content_uri = None
        if not prune:
            content_uri = pnts_path.relative_to(folder)
            xyz_float = xyz.view(np.float32).reshape((fth.points_length, 3))

            # update aabb based on real values
            bounding_box = BoundingVolumeBox.from_points(xyz_float)

        else:
            # if it is a leaf that should be pruned
            if not children_tileset_parts:
                return None

            # recompute the aabb in function of children
            bounding_box = BoundingVolumeBox()
            for child_tileset_part in children_tileset_parts:
                if child_tileset_part.bounding_volume is not None:
                    bounding_box.add(child_tileset_part.bounding_volume)

        if bounding_box is None:
            raise TilerException("bounding_box shouldn't be None")

        tile: Tile = Tile(
            geometric_error=10 * self.spacing / scale[0], bounding_volume=bounding_box
        )
        if content_uri is not None:
            tile.content_uri = content_uri

        if children_tileset_parts:
            tile.children = children_tileset_parts
        else:
            tile.geometric_error = 0.0

        if (
            len(self.name) > 0
            and children_tileset_parts
            and len(json.dumps(tile.to_dict())) > 100000
        ):
            tile = split_tileset(tile, self.name.decode(), folder)

        return tile


def split_tileset(tile: Tile, split_name: str, folder: Path) -> Tile:
    tile.set_refine_mode("ADD")
    tileset = TileSet(geometric_error=tile.geometric_error)
    tileset.root_tile = copy.deepcopy(tile)
    tileset_name = Path(f"tileset.{split_name}.json")
    tileset.write_as_json(folder / tileset_name)
    tile.content_uri = tileset_name
    tile.children = []

    return tile



import pickle
import time
from typing import Generator, List, Optional, TextIO, Tuple

from py3dtiles.tilers.point.node.node import Node
from py3dtiles.tilers.point.node.node_catalog import NodeCatalog


class NodeProcess:
    def __init__(
        self,
        node_catalog: NodeCatalog,
        scale: float,
        name: bytes,
        tasks: List[bytes],
        begin: float,
        log_file: Optional[TextIO],
    ):
        self.node_catalog = node_catalog
        self.scale = scale
        self.name = name
        self.tasks = tasks
        self.begin = begin
        self.log_file = log_file
        self.total_point_count = 0

    def _flush(
        self,
        node: Node,
        max_depth: int = 1,
        force_forward: bool = False,
        depth: int = 0,
    ) -> Generator[Tuple[bytes, bytes, int], None, None]:
        if depth >= max_depth:
            threshold = 0 if force_forward else 10_000
            if node.get_pending_points_count() > threshold:
                yield from node.dump_pending_points()
            return

        node.flush_pending_points(self.node_catalog, self.scale)
        if node.children is not None:
            # then flush children
            children = node.children
            # release node
            del node
            for child_name in children:
                yield from self._flush(
                    self.node_catalog.get_node(child_name),
                    max_depth,
                    force_forward,
                    depth + 1,
                )

    def _balance(self, node: Node, max_depth: int = 1, depth: int = 0) -> None:
        if depth >= max_depth:
            return

        if node.needs_balance():
            node.grid.balance(node.aabb_size, node.aabb[0], node.inv_aabb_size)
            node.dirty = True

        if node.children is not None:
            # then _balance children
            children = node.children
            # release node
            del node
            for child_name in children:
                self._balance(
                    self.node_catalog.get_node(child_name), max_depth, depth + 1
                )

    def infer_depth_from_name(self) -> int:
        halt_at_depth = 0
        if len(self.name) >= 7:
            halt_at_depth = 5
        elif len(self.name) >= 5:
            halt_at_depth = 3
        elif len(self.name) > 2:
            halt_at_depth = 2
        elif len(self.name) >= 1:
            halt_at_depth = 1
        return halt_at_depth

    def run(self) -> Generator[Tuple[bytes, bytes, int], None, None]:
        log_enabled = self.log_file is not None

        if log_enabled:
            print(
                f'[>] process_node: "{self.name!r}", {len(self.tasks)}',
                file=self.log_file,
                flush=True,
            )

        node = self.node_catalog.get_node(self.name)

        halt_at_depth = self.infer_depth_from_name()

        for index, task in enumerate(self.tasks):
            if log_enabled:
                print(
                    f"  -> read source [{time.time() - self.begin}]",
                    file=self.log_file,
                    flush=True,
                )

            data = pickle.loads(task)

            point_count = len(data["xyz"])

            if log_enabled:
                print(
                    "  -> insert {} [{} points]/ {} files [{}]".format(
                        index + 1,
                        point_count,
                        len(self.tasks),
                        time.time() - self.begin,
                    ),
                    file=self.log_file,
                    flush=True,
                )

            # insert points in node (no children handling here)
            node.insert(
                self.scale,
                data["xyz"],
                data["rgb"],
                data["classification"],
                data["intensity"],
                halt_at_depth == 0,
            )

            self.total_point_count += point_count

            if log_enabled:
                print(
                    f"  -> _flush [{time.time() - self.begin}]",
                    file=self.log_file,
                    flush=True,
                )
            # _flush push pending points (= call insert) from level N to level N + 1
            # (_flush is recursive)
            for flush_name, flush_data, flush_point_count in self._flush(
                node,
                halt_at_depth - 1,
                index == len(self.tasks) - 1,
            ):
                self.total_point_count -= flush_point_count
                yield flush_name, flush_data, flush_point_count

        self._balance(node, halt_at_depth - 1)



from pathlib import Path

from numpy.testing import assert_array_almost_equal, assert_array_equal

from py3dtiles.merger import merge, merge_from_files, merge_with_pnts_content
from py3dtiles.tileset import BoundingVolumeBox, TileSet
from py3dtiles.tileset.content import Pnts, read_binary_tile_content


def test_merge_with_memory_tilesets(
    tmp_dir: Path, tileset_1: TileSet, tileset_2: TileSet
) -> None:
    merged_tileset = merge([tileset_1, tileset_2])

    assert len(merged_tileset.root_tile.get_all_children()) == 2

    # The main tilesets should link the tilesets
    assert merged_tileset.root_tile.children[0].tile_content is tileset_1
    assert merged_tileset.root_tile.children[1].tile_content is tileset_2

    # check if there is no URI
    for tile_child in merged_tileset.root_tile.get_all_children():
        assert tile_child.content_uri is None

    assert merged_tileset.geometric_error == max(
        tileset_1.root_tile.geometric_error, tileset_2.root_tile.geometric_error
    )
    assert merged_tileset.root_tile.geometric_error == max(
        tileset_1.root_tile.geometric_error, tileset_2.root_tile.geometric_error
    )

    # check if the bounding boxes are consistent
    for tile_child in merged_tileset.root_tile.get_all_children():
        sub_tileset = tile_child.get_or_fetch_content(tmp_dir)
        assert isinstance(sub_tileset, TileSet)

        sub_root_bounding_box = sub_tileset.root_tile.bounding_volume
        assert isinstance(sub_root_bounding_box, BoundingVolumeBox)
        sub_root_bounding_box.transform(sub_tileset.root_tile.transform)

        child_bounding_box = tile_child.bounding_volume
        assert isinstance(child_bounding_box, BoundingVolumeBox)
        child_bounding_box.transform(
            merged_tileset.root_tile.transform @ tile_child.transform
        )

        assert_array_almost_equal(sub_root_bounding_box._box, child_bounding_box._box)  # type: ignore [arg-type]

    assert merged_tileset.root_tile.content_uri is None
    assert merged_tileset.root_tile.tile_content is None


def test_merge_with_pnts_content_with_memory_tilesets(
    tmp_dir: Path, tileset_1: TileSet, tileset_2: TileSet
) -> None:
    merged_tileset = merge_with_pnts_content([tileset_1, tileset_2])

    assert len(merged_tileset.root_tile.get_all_children()) == 2

    # The main tilesets should link the tilesets
    assert merged_tileset.root_tile.children[0].tile_content is tileset_1
    assert merged_tileset.root_tile.children[1].tile_content is tileset_2

    # check if there is no URI
    for tile_child in merged_tileset.root_tile.get_all_children():
        assert tile_child.content_uri is None

    assert merged_tileset.geometric_error == 1720.2520618394917
    assert merged_tileset.root_tile.geometric_error == 137.61986044319053

    # check if the bounding boxes are consistent
    for tile_child in merged_tileset.root_tile.get_all_children():
        sub_tileset = tile_child.get_or_fetch_content(tmp_dir)
        assert isinstance(sub_tileset, TileSet)

        sub_root_bounding_box = sub_tileset.root_tile.bounding_volume
        assert isinstance(sub_root_bounding_box, BoundingVolumeBox)
        sub_root_bounding_box.transform(sub_tileset.root_tile.transform)

        child_bounding_box = tile_child.bounding_volume
        assert isinstance(child_bounding_box, BoundingVolumeBox)
        child_bounding_box.transform(
            merged_tileset.root_tile.transform @ tile_child.transform
        )

        assert_array_almost_equal(sub_root_bounding_box._box, child_bounding_box._box)  # type: ignore [arg-type]

    assert merged_tileset.root_tile.content_uri == Path("r.pnts")

    merged_pnts = merged_tileset.root_tile.tile_content
    assert isinstance(merged_pnts, Pnts)

    assert merged_pnts.body.feature_table.nb_points() == 790


def test_merge_with_file_tilesets(
    tmp_dir: Path, tileset_path_1: Path, tileset_path_2: Path
) -> None:
    tileset_1 = TileSet.from_file(tileset_path_1)
    tileset_2 = TileSet.from_file(tileset_path_2)

    merged_tileset_path = tmp_dir / "merged_tileset.json"

    merge_from_files(
        [tileset_path_1, tileset_path_2],
        merged_tileset_path,
        overwrite=False,
        force_universal_merger=True,
    )

    # The main tilesets should not be modified (for the moment)
    assert tileset_1.to_dict() == TileSet.from_file(tileset_path_1).to_dict()
    assert tileset_2.to_dict() == TileSet.from_file(tileset_path_2).to_dict()

    merged_tileset = TileSet.from_file(merged_tileset_path)

    assert len(merged_tileset.root_tile.get_all_children()) == 2

    # check if URIs are present
    for result, expected in zip(
        merged_tileset.root_tile.get_all_children(), (tileset_path_1, tileset_path_2)
    ):
        assert result.content_uri == expected.relative_to(merged_tileset_path.parent)

    assert merged_tileset.geometric_error == max(
        tileset_1.root_tile.geometric_error, tileset_2.root_tile.geometric_error
    )
    assert merged_tileset.root_tile.geometric_error == max(
        tileset_1.root_tile.geometric_error, tileset_2.root_tile.geometric_error
    )

    # check if the bounding boxes are consistent
    for tile_child in merged_tileset.root_tile.get_all_children():
        sub_tileset = tile_child.get_or_fetch_content(tmp_dir)
        assert isinstance(sub_tileset, TileSet)

        sub_root_bounding_box = sub_tileset.root_tile.bounding_volume
        assert isinstance(sub_root_bounding_box, BoundingVolumeBox)
        sub_root_bounding_box.transform(sub_tileset.root_tile.transform)

        child_bounding_box = tile_child.bounding_volume
        assert isinstance(child_bounding_box, BoundingVolumeBox)
        child_bounding_box.transform(
            merged_tileset.root_tile.transform @ tile_child.transform
        )

        assert_array_equal(sub_root_bounding_box._box, child_bounding_box._box)  # type: ignore [arg-type]

    assert merged_tileset.root_tile.content_uri is None
    assert merged_tileset.root_tile.tile_content is None

    merged_tileset_path.unlink()


def test_merge_with_pnts_content_with_file_tilesets(
    tmp_dir: Path, tileset_path_1: Path, tileset_path_2: Path
) -> None:
    tileset_1 = TileSet.from_file(tileset_path_1)
    tileset_2 = TileSet.from_file(tileset_path_2)

    merged_tileset_path = tmp_dir / "merged_tileset.json"

    merge_from_files(
        [tileset_path_1, tileset_path_2],
        merged_tileset_path,
        overwrite=False,
    )

    # The main tilesets should not be modified (for the moment)
    assert tileset_1.to_dict() == TileSet.from_file(tileset_path_1).to_dict()
    assert tileset_2.to_dict() == TileSet.from_file(tileset_path_2).to_dict()

    merged_tileset = TileSet.from_file(merged_tileset_path)

    assert len(merged_tileset.root_tile.get_all_children()) == 2

    # check if URIs are present
    for result, expected in zip(
        merged_tileset.root_tile.get_all_children(), (tileset_path_1, tileset_path_2)
    ):
        assert result.content_uri == expected.relative_to(merged_tileset_path.parent)

    assert merged_tileset.geometric_error == 1720.2520618394917
    assert merged_tileset.root_tile.geometric_error == 137.61986044319053

    # check if the bounding boxes are consistent
    for tile_child in merged_tileset.root_tile.get_all_children():
        sub_tileset = tile_child.get_or_fetch_content(tmp_dir)
        assert isinstance(sub_tileset, TileSet)

        sub_root_bounding_box = sub_tileset.root_tile.bounding_volume
        assert isinstance(sub_root_bounding_box, BoundingVolumeBox)
        sub_root_bounding_box.transform(sub_tileset.root_tile.transform)

        child_bounding_box = tile_child.bounding_volume
        assert isinstance(child_bounding_box, BoundingVolumeBox)
        child_bounding_box.transform(
            merged_tileset.root_tile.transform @ tile_child.transform
        )

        assert_array_almost_equal(sub_root_bounding_box._box, child_bounding_box._box)  # type: ignore [arg-type]

    merged_pnts_path = merged_tileset_path.parent / "r.pnts"
    assert merged_pnts_path.exists()

    merged_pnts = read_binary_tile_content(merged_pnts_path)
    assert isinstance(merged_pnts, Pnts)

    assert merged_pnts.body.feature_table.nb_points() == 790



import concurrent.futures
import pickle
import struct
import time
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
from pyproj import CRS, Transformer

from py3dtiles.exceptions import (
    SrsInMissingException,
    SrsInMixinException,
    TilerException,
)
from py3dtiles.tilers.base_tiler import Tiler
from py3dtiles.tileset.content import read_binary_tile_content
from py3dtiles.tileset.tileset import TileSet
from py3dtiles.utils import (
    READER_MAP,
    compute_spacing,
    make_aabb_valid,
    node_from_name,
    node_name_to_path,
)

from .matrix_manipulation import (
    make_rotation_matrix,
    make_scale_matrix,
    make_translation_matrix,
)
from .node import Node, SharedNodeStore
from .pnts import MIN_POINT_SIZE, pnts_writer
from .point_message_type import PointManagerMessage, PointWorkerMessageType
from .point_shared_metadata import PointSharedMetadata
from .point_state import PointState
from .point_tiler_worker import PointTilerWorker


def is_ancestor(node_name: bytes, ancestor: bytes) -> bool:
    """
    Example, the tile 22 is ancestor of 22458
    Particular case, the tile 22 is ancestor of 22
    """
    return len(ancestor) <= len(node_name) and node_name[0 : len(ancestor)] == ancestor


def is_ancestor_in_list(
    node_name: bytes, ancestors: Union[List[bytes], Dict[bytes, Any]]
) -> bool:
    return any(
        not ancestor or is_ancestor(node_name, ancestor) for ancestor in ancestors
    )


def can_pnts_be_written(
    node_name: bytes,
    finished_node: bytes,
    input_nodes: Union[List[bytes], Dict[bytes, Any]],
    active_nodes: Union[List[bytes], Dict[bytes, Any]],
) -> bool:
    return (
        is_ancestor(node_name, finished_node)
        and not is_ancestor_in_list(node_name, active_nodes)
        and not is_ancestor_in_list(node_name, input_nodes)
    )


class PointTiler(Tiler[PointSharedMetadata, PointTilerWorker]):
    name = b"points"

    file_info: Dict[str, Any]
    root_aabb: npt.NDArray[np.float64]
    root_scale: npt.NDArray[np.float32]
    root_spacing: float
    node_store: SharedNodeStore
    state: PointState

    def __init__(
        self,
        out_folder: Path,
        files: Union[List[Union[str, Path]], str, Path],
        crs_in: Optional[CRS],
        force_crs_in: bool,
        rgb: bool,
        classification: bool,
        intensity: bool,
        color_scale: Optional[float],
        cache_size: int,
        verbosity: int,
    ):
        self.out_folder = out_folder

        # allow str directly if only one input
        files = [files] if isinstance(files, (str, Path)) else files
        self.files = [Path(file) for file in files]

        self.rgb = rgb
        self.classification = classification
        self.intensity = intensity
        self.color_scale = color_scale

        self.crs_in = crs_in
        self.force_crs_in = force_crs_in

        self.cache_size = cache_size

        self.verbosity = verbosity

    def get_worker(self) -> PointTilerWorker:
        return PointTilerWorker(self.shared_metadata)

    def get_tasks(
        self, startup: float
    ) -> Generator[Tuple[bytes, List[bytes]], None, None]:
        while len(self.state.pnts_to_writing) > 0:
            yield self.send_pnts_to_write()

        yield from self.send_points_to_process(time.time() - startup)

        while self.state.can_add_reading_jobs():
            yield self.send_file_to_read()

    def initialization(
        self,
        crs_out: Optional[CRS],
        working_dir: Path,
        number_of_jobs: int,
    ) -> None:
        self.file_info = self.get_file_info(self.crs_in, self.force_crs_in)
        transformer = self.get_transformer(crs_out)
        (
            self.rotation_matrix,
            self.original_aabb,
            self.avg_min,
        ) = self.get_rotation_matrix(crs_out, transformer)

        self.root_aabb, self.root_scale, self.root_spacing = self.get_root_aabb(
            self.original_aabb
        )

        self.node_store = SharedNodeStore(working_dir)

        self.state = PointState(self.file_info["portions"], max(1, number_of_jobs // 2))

        self.shared_metadata = PointSharedMetadata(
            transformer,
            self.root_aabb,
            self.root_spacing,
            self.root_scale,
            self.out_folder,
            self.rgb,
            self.color_scale,
            self.classification,
            self.intensity,
            self.verbosity,
        )

    def get_file_info(
        self,
        crs_in: Optional[CRS],
        force_crs_in: bool = False,
    ) -> Dict[str, Any]:

        pointcloud_file_portions = []
        aabb = None
        total_point_count = 0
        avg_min = np.array([0.0, 0.0, 0.0])

        # read all input files headers and determine the aabb/spacing
        for file in self.files:
            extension = file.suffix
            if extension in READER_MAP:
                reader = READER_MAP[extension]
            else:
                raise ValueError(
                    f"The file with {extension} extension can't be read, "
                    f"the available extensions are: {READER_MAP.keys()}"
                )

            file_info = reader.get_metadata(file)

            pointcloud_file_portions += file_info["portions"]
            if aabb is None:
                aabb = file_info["aabb"]
            else:
                aabb[0] = np.minimum(aabb[0], file_info["aabb"][0])
                aabb[1] = np.maximum(aabb[1], file_info["aabb"][1])

            file_crs_in = file_info["crs_in"]
            if file_crs_in is not None:
                if crs_in is None:
                    crs_in = file_crs_in
                elif crs_in != file_crs_in and not force_crs_in:
                    raise SrsInMixinException(
                        "All input files should have the same srs in, currently there are a mix of"
                        f" {crs_in} and {file_crs_in}"
                    )
            total_point_count += file_info["point_count"]
            avg_min += file_info["avg_min"] / len(self.files)

        # The fact self.files is not empty have been checked before, so this shouldn't happen
        # but this keeps mypy happy and also serve as "defensive programming"
        if aabb is None:
            raise RuntimeError("No aabb could be computed!")
        # correct aabb, so that we don't have null sized box
        # we add 10^-5, supposing it's reasonable for most use case
        make_aabb_valid(aabb)
        return {
            "portions": pointcloud_file_portions,
            "aabb": aabb,
            "crs_in": crs_in,
            "point_count": total_point_count,
            "avg_min": avg_min,
        }

    def get_transformer(self, crs_out: Optional[CRS]) -> Optional[Transformer]:
        if crs_out:
            if self.file_info["crs_in"] is None:
                raise SrsInMissingException(
                    "None file has a input srs specified. Should be provided."
                )

            transformer = Transformer.from_crs(self.file_info["crs_in"], crs_out)
        else:
            transformer = None

        return transformer

    def get_rotation_matrix(
        self, crs_out: Optional[CRS], transformer: Optional[Transformer]
    ) -> Tuple[
        npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]
    ]:
        avg_min: npt.NDArray[np.float64] = self.file_info["avg_min"]
        aabb: npt.NDArray[np.float64] = self.file_info["aabb"]

        rotation_matrix: npt.NDArray[np.float64] = np.identity(4)
        if crs_out is not None and transformer is not None:

            bl: npt.NDArray[np.float64] = np.array(
                list(transformer.transform(aabb[0][0], aabb[0][1], aabb[0][2]))
            )
            tr: npt.NDArray[np.float64] = np.array(
                list(transformer.transform(aabb[1][0], aabb[1][1], aabb[1][2]))
            )
            br: npt.NDArray[np.float64] = np.array(
                list(transformer.transform(aabb[1][0], aabb[0][1], aabb[0][2]))
            )

            avg_min = np.array(
                list(transformer.transform(avg_min[0], avg_min[1], avg_min[2]))
            )

            x_axis = br - bl

            bl = bl - avg_min
            tr = tr - avg_min

            if crs_out.to_epsg() == 4978:
                # Transform geocentric normal => (0, 0, 1)
                # and 4978-bbox x axis => (1, 0, 0),
                # to have a bbox in local coordinates that's nicely aligned with the data
                rotation_matrix = make_rotation_matrix(avg_min, np.array([0, 0, 1]))
                rotation_matrix = np.dot(
                    make_rotation_matrix(x_axis, np.array([1, 0, 0])), rotation_matrix
                )

                rotation_matrix_part = rotation_matrix[:3, :3].T

                bl = np.dot(bl, rotation_matrix_part)
                tr = np.dot(tr, rotation_matrix_part)

            root_aabb = np.array([np.minimum(bl, tr), np.maximum(bl, tr)])
        else:
            # offset
            root_aabb = aabb - avg_min

        return rotation_matrix, root_aabb, avg_min

    def get_root_aabb(
        self, original_aabb: npt.NDArray[np.float64]
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float32], float]:
        base_spacing = compute_spacing(original_aabb)
        if base_spacing > 10:
            root_scale = np.array([0.01, 0.01, 0.01])
        elif base_spacing > 1:
            root_scale = np.array([0.1, 0.1, 0.1])
        else:
            root_scale = np.array([1, 1, 1])

        root_aabb = original_aabb * root_scale
        root_spacing = compute_spacing(root_aabb)
        return root_aabb, root_scale, root_spacing

    def print_summary(self) -> None:
        print("Summary:")
        print("  - points to process: {}".format(self.file_info["point_count"]))
        print(f"  - offset to use: {self.avg_min}")
        print(f"  - root spacing: {self.root_spacing / self.root_scale[0]}")
        print(f"  - root aabb: {self.root_aabb}")
        print(f"  - original aabb: {self.original_aabb}")
        print(f"  - scale: {self.root_scale}")

    def send_file_to_read(self) -> Tuple[bytes, List[bytes]]:
        if self.verbosity >= 1:
            print(f"Submit next portion {self.state.point_cloud_file_parts[-1]}")
        file, portion = self.state.point_cloud_file_parts.pop()
        self.state.points_in_progress += portion[1] - portion[0]

        self.state.number_of_reading_jobs += 1

        return PointManagerMessage.READ_FILE.value, [
            pickle.dumps(
                {
                    "filename": file,
                    "offset_scale": (
                        -self.avg_min,
                        self.root_scale,
                        self.rotation_matrix[:3, :3].T,
                    ),
                    "portion": portion,
                }
            ),
        ]

    def send_points_to_process(
        self, now: float
    ) -> Generator[Tuple[bytes, List[bytes]], None, None]:
        potentials = sorted(
            # a key (=task) can be in node_to_process and processing_nodes if the node isn't completely processed
            [
                (node, task)
                for node, task in self.state.node_to_process.items()  # task: [data...], point_count
                if node not in self.state.processing_nodes
            ],
            key=lambda task: -len(task[0]),
        )  # sort by node name size, the root nodes first

        while potentials:
            target_count = 100_000
            job_list = []
            count = 0
            idx = len(potentials) - 1
            while count < target_count and idx >= 0:
                name, (tasks, point_count) = potentials[idx]
                count += point_count
                job_list += [
                    name,
                    self.node_store.get(name),
                    struct.pack(">I", len(tasks)),
                ] + tasks
                del potentials[idx]

                del self.state.node_to_process[name]
                self.state.processing_nodes[name] = (
                    len(tasks),
                    point_count,
                    now,
                )

                if name in self.state.waiting_writing_nodes:
                    self.state.waiting_writing_nodes.pop(
                        self.state.waiting_writing_nodes.index(name)
                    )
                idx -= 1

            if job_list:
                yield PointManagerMessage.PROCESS_JOBS.value, job_list

    def send_pnts_to_write(self) -> Tuple[bytes, List[bytes]]:
        node_name = self.state.pnts_to_writing.pop()
        data = self.node_store.get(node_name)
        if not data:
            raise ValueError(f"{node_name!r} has no data")

        self.node_store.remove(node_name)
        self.state.number_of_writing_jobs += 1

        return PointManagerMessage.WRITE_PNTS.value, [node_name, data]

    def process_message(self, return_type: bytes, result: List[bytes]) -> bool:
        at_least_one_job_ended = False

        if return_type == PointWorkerMessageType.READ.value:
            self.state.number_of_reading_jobs -= 1
            at_least_one_job_ended = True

        elif return_type == PointWorkerMessageType.PROCESSED.value:
            content = pickle.loads(result[-1])
            self.state.processed_points += content["total"]
            self.state.points_in_progress -= content["total"]

            del self.state.processing_nodes[content["name"]]

            self.dispatch_processed_nodes(content)

            at_least_one_job_ended = True

        elif return_type == PointWorkerMessageType.PNTS_WRITTEN.value:
            self.state.points_in_pnts += struct.unpack(">I", result[0])[0]
            self.state.number_of_writing_jobs -= 1

        elif return_type == PointWorkerMessageType.NEW_TASK.value:
            self.state.add_tasks_to_process(
                node_name=result[0],
                data=result[1],
                point_count=struct.unpack(">I", result[2])[0],
            )

        else:
            raise NotImplementedError(f"The command {return_type!r} is not implemented")

        return at_least_one_job_ended

    def dispatch_processed_nodes(self, content: Dict[str, bytes]) -> None:
        if not content["name"]:
            return

        self.node_store.put(content["name"], content["data"])
        self.state.waiting_writing_nodes.append(content["name"])

        if not self.state.is_reading_finish():
            return

        # if all nodes aren't processed yet,
        # we should check if linked ancestors are processed
        if self.state.processing_nodes or self.state.node_to_process:
            finished_node = content["name"]
            if can_pnts_be_written(
                finished_node,
                finished_node,
                self.state.node_to_process,
                self.state.processing_nodes,
            ):
                self.state.waiting_writing_nodes.pop(-1)
                self.state.pnts_to_writing.append(finished_node)

                for i in range(len(self.state.waiting_writing_nodes) - 1, -1, -1):
                    candidate = self.state.waiting_writing_nodes[i]

                    if can_pnts_be_written(
                        candidate,
                        finished_node,
                        self.state.node_to_process,
                        self.state.processing_nodes,
                    ):
                        self.state.waiting_writing_nodes.pop(i)
                        self.state.pnts_to_writing.append(candidate)

        else:
            for c in self.state.waiting_writing_nodes:
                self.state.pnts_to_writing.append(c)
            self.state.waiting_writing_nodes.clear()

    def validate_binary_data(self) -> None:
        if self.state.points_in_pnts != self.file_info["point_count"]:
            raise ValueError(
                "Invalid point count in the written .pnts"
                + f"(expected: {self.file_info['point_count']}, was: {self.state.points_in_pnts})"
            )

    def write_tileset(self) -> None:
        # compute tile transform matrix
        transform = np.linalg.inv(self.rotation_matrix)
        transform = np.dot(transform, make_scale_matrix(1.0 / self.root_scale[0]))
        transform = np.dot(make_translation_matrix(self.avg_min), transform)

        # Create the root tile by sampling (or taking all points?) of child nodes
        root_node = Node(b"", self.root_aabb, self.root_spacing * 2)
        root_node.children = []
        inv_aabb_size = (
            1.0
            / np.maximum(
                MIN_POINT_SIZE,
                self.root_aabb[1] - self.root_aabb[0],
            )
        ).astype(np.float32)
        for child_num in range(8):
            tile_path = node_name_to_path(
                self.out_folder, str(child_num).encode("ascii"), ".pnts"
            )
            if tile_path.exists():
                tile_content = read_binary_tile_content(tile_path)

                fth = tile_content.body.feature_table.header
                xyz = tile_content.body.feature_table.body.position.view(
                    np.float32
                ).reshape((fth.points_length, 3))
                if self.rgb:
                    tile_color = tile_content.body.feature_table.body.color
                    if tile_color is None:
                        raise TilerException(
                            "tile_content.body.feature_table.body.color shouldn't be None here. Seems to be a py3dtiles issue."
                        )
                    if tile_color.dtype != np.uint8:
                        raise TilerException(
                            "The data type of tile_content.body.feature_table.body.color must be np.uint8. Seems to be a py3dtiles issue."
                        )
                    rgb = tile_color.reshape((fth.points_length, 3)).astype(
                        np.uint8, copy=False
                    )  # the astype is used for typing
                else:
                    rgb = np.zeros(xyz.shape, dtype=np.uint8)
                if self.classification:
                    classification = (
                        tile_content.body.batch_table.get_binary_property(
                            "Classification"
                        )
                        .astype(np.uint8)
                        .reshape(-1, 1)
                    )
                else:
                    classification = np.zeros((fth.points_length, 1), dtype=np.uint8)

                if self.intensity:
                    intensity = (
                        (tile_content.body.batch_table.get_binary_property("Intensity"))
                        .astype(np.uint8)
                        .reshape(-1, 1)
                    )
                else:
                    intensity = np.zeros((fth.points_length, 1), dtype=np.uint8)

                root_node.grid.insert(
                    self.root_aabb[0].astype(np.float32),
                    inv_aabb_size,
                    xyz.copy(),
                    rgb,
                    classification,
                    intensity,
                )

        pnts_writer.node_to_pnts(
            b"",
            root_node,
            self.out_folder,
            self.rgb,
            self.classification,
            self.intensity,
        )

        pool_executor = concurrent.futures.ProcessPoolExecutor()
        root_tile = node_from_name(b"", self.root_aabb, self.root_spacing).to_tileset(
            self.out_folder, self.root_scale, None, 0, pool_executor
        )
        pool_executor.shutdown()

        if root_tile is None:
            raise RuntimeError(
                "root_tileset cannot be None here. This is likely a tiler bug."
            )

        root_tile.transform = transform.reshape(16, order="F")
        root_tile.set_refine_mode(
            "REPLACE"
        )  # The root tile is in the "REPLACE" refine mode
        # And children with the "ADD" refine mode
        # No need to set this property in their children, they will take the parent value if it is not present
        for child in root_tile.children:
            child.set_refine_mode("ADD")

        geometric_error = (
            np.linalg.norm(self.root_aabb[1] - self.root_aabb[0]) / self.root_scale[0]
        )
        tileset = TileSet(geometric_error=geometric_error)
        tileset.root_tile = root_tile
        tileset.write_as_json(self.out_folder / "tileset.json")

    def benchmark(self, benchmark_id: str, startup: float) -> None:
        print(
            "{},{},{},{}".format(
                self.benchmark,
                ",".join([f.name for f in self.files]),
                self.state.points_in_pnts,
                round(time.time() - startup, 1),
            )
        )

    def print_debug(
        self, now: float, number_of_jobs: int, number_of_idle_clients: int
    ) -> None:
        if self.verbosity >= 3:
            print("{:^16}|{:^8}|{:^8}".format("Name", "Points", "Seconds"))
            for name, v in self.state.processing_nodes.items():
                print(
                    "{:^16}|{:^8}|{:^8}".format(
                        "{} ({})".format(name.decode("ascii"), v[0]),
                        v[1],
                        round(now - v[2], 1),
                    )
                )
            print("")
            print("Pending:")
            print(
                "  - root: {} / {}".format(
                    len(self.state.point_cloud_file_parts),
                    self.state.initial_portion_count,
                )
            )
            print(
                "  - other: {} files for {} nodes".format(
                    sum([len(f[0]) for f in self.state.node_to_process.values()]),
                    len(self.state.node_to_process),
                )
            )
            print("")

        elif self.verbosity >= 2:
            self.state.print_debug()

        if self.verbosity >= 1:
            print(
                "{} % points in {} sec [{} tasks, {} nodes, {} wip]".format(
                    round(
                        100
                        * self.state.processed_points
                        / self.file_info["point_count"],
                        2,
                    ),
                    round(now, 1),
                    number_of_jobs - number_of_idle_clients,
                    len(self.state.processing_nodes),
                    self.state.points_in_progress,
                )
            )

        elif self.verbosity >= 0:
            percent = round(
                100 * self.state.processed_points / self.file_info["point_count"],
                2,
            )
            time_left = (100 - percent) * now / (percent + 0.001)
            print(
                f"\r{percent:>6} % in {round(now)} sec [est. time left: {round(time_left)} sec]      ",
                end="",
                flush=True,
            )

    def memory_control(self) -> None:
        self.node_store.control_memory_usage(self.cache_size, self.verbosity)



import json
from pathlib import Path

from py3dtiles.tileset.extension.batch_table_hierarchy_extension import (
    BatchTableHierarchy,
)


def test_bth_build_sample_with_indexes_and_compare_reference_file(
    batch_table_hierarchy_with_indexes: BatchTableHierarchy,
    batch_table_hierarchy_reference_file: Path,
) -> None:
    """
    Build the sample, load the version from the reference file and
    compare them (in memory as opposed to "diffing" files)
    """
    json_bth = batch_table_hierarchy_with_indexes.to_dict()

    with batch_table_hierarchy_reference_file.open() as reference_file:
        json_reference = json.loads(reference_file.read())

    json_reference.pop("_comment", None)  # Drop the "comment".

    assert json_bth == json_reference


def test_bth_build_sample_with_instances_and_compare_reference_file(
    batch_table_hierarchy_with_instances: BatchTableHierarchy,
    batch_table_hierarchy_reference_file: Path,
) -> None:
    """
    Build the sample, load the version from the reference file and
    compare them (in memory as opposed to "diffing" files)
    """
    json_bth = batch_table_hierarchy_with_instances.to_dict()

    with batch_table_hierarchy_reference_file.open() as reference_file:
        json_reference = json.loads(reference_file.read())

    json_reference.pop("_comment", None)  # Drop the "comment".

    assert json_bth == json_reference



import doctest
import shutil
from pathlib import Path

if __name__ == "__main__":
    # With doctest.ELLIPSIS, an ellipsis marker (...) in the expected output can match any substring in the actual output.
    # Useful to match strings like "<py3dtiles.tileset.content.pnts.PntsHeader object at 0x7f73f5530d90>"
    test_result = doctest.testfile("../docs/api.rst", optionflags=doctest.ELLIPSIS)
    num_of_failed = test_result.failed
    test_result = doctest.testfile("../README.rst", optionflags=doctest.ELLIPSIS)
    num_of_failed += test_result.failed

    # Remove files created by the tested files.
    Path("mymodel.b3dm").unlink(missing_ok=True)
    Path("mypoints.pnts").unlink(missing_ok=True)
    shutil.rmtree("./3dtiles_output", ignore_errors=True)
    shutil.rmtree("./my3dtiles", ignore_errors=True)
    shutil.rmtree("./my3dtiles2", ignore_errors=True)

    exit(num_of_failed != 0)



"""
This module contains supporting classes and functions to manipulate 3DTiles data structure.

Use this module if you plan on directly reading, manipulating and writing tilesets, tiles and tile contents.
"""
from .bounding_volume import BoundingVolume
from .bounding_volume_box import BoundingVolumeBox
from .root_property import RootProperty
from .tile import Tile
from .tileset import TileSet
from .utils import number_of_points_in_tileset

__all__ = [
    "BoundingVolume",
    "BoundingVolumeBox",
    "content",
    "extension",
    "number_of_points_in_tileset",
    "RootProperty",
    "Tile",
    "TileSet",
]



from __future__ import annotations

from typing import Any

from py3dtiles.tileset.extension.base_extension import BaseExtension
from py3dtiles.typing import ExtensionDictType, HierarchyClassDictType


class HierarchyInstance:
    def __init__(
        self,
        properties: dict[str, Any] | None = None,
        parents: list[int | HierarchyInstance] | None = None,
    ):
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}

        if parents is not None:
            self.parents = parents
        else:
            self.parents = []


class HierarchyClass:
    def __init__(self, name: str, property_names: list[str]) -> None:
        self.name = name
        self.instances: list[HierarchyInstance] = []
        self.property_names = property_names

    def add_instance(
        self,
        properties: dict[str, Any],
        parents: list[int | HierarchyInstance] | None = None,
    ) -> HierarchyInstance:
        hierarchy_instance = HierarchyInstance(parents=parents)
        for name in self.property_names:
            if name in properties:
                hierarchy_instance.properties[name] = properties[name]
        self.instances.append(hierarchy_instance)
        return hierarchy_instance

    def to_dict(self) -> HierarchyClassDictType:
        return {
            "name": self.name,
            "length": len(self.instances),
            "instances": {
                name: [instance.properties[name] for instance in self.instances]
                for name in self.property_names
            },
        }


class BatchTableHierarchy(BaseExtension):
    """
    Batch Table Hierarchy (BTH) is a BaseExtension of a Batch Table.
    """

    def __init__(self) -> None:
        super().__init__("3DTILES_batch_table_hierarchy")
        self.classes: list[HierarchyClass] = []
        self.instancesLength = 0

    @classmethod
    def from_dict(cls, extension_dict: ExtensionDictType) -> BatchTableHierarchy:
        raise NotImplementedError(
            "The load from a dictionary of the BatchTableHierarchy is not implemented."
        )

    def add_class(self, class_name: str, property_names: list[str]) -> HierarchyClass:
        hierarchy_class = HierarchyClass(class_name, property_names)
        self.classes.append(hierarchy_class)
        return hierarchy_class

    def get_instance_parent_indexes(self, instance: HierarchyInstance) -> list[int]:
        parent_indexes = []
        for parent in instance.parents:
            if isinstance(parent, int):
                parent_indexes.append(parent)
            else:
                index = 0
                for hierarchy_class in self.classes:
                    for hierarchy_instance in hierarchy_class.instances:
                        if parent == hierarchy_instance:
                            parent_indexes.append(index)
                        index += 1
        return parent_indexes

    def to_dict(self) -> ExtensionDictType:
        dict_data: ExtensionDictType = {
            "classes": [],
            "classIds": [],
            "parentCounts": [],
            "parentIds": [],
            "instancesLength": 0,
        }

        for class_id, hierarchy_class in enumerate(self.classes):
            dict_data["classes"].append(hierarchy_class.to_dict())
            for instance in hierarchy_class.instances:
                dict_data["classIds"].append(class_id)
                dict_data["parentCounts"].append(len(instance.parents))
                dict_data["parentIds"].extend(
                    self.get_instance_parent_indexes(instance)
                )
            dict_data["instancesLength"] += len(hierarchy_class.instances)

        return dict_data



import math
import unittest
from typing import List

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal

from py3dtiles.tileset import BoundingVolumeBox

# fmt: off
DUMMY_MATRIX = [
    1, 2, 3, 4,
    5, 6, 7, 8,
    9, 10, 11, 12,
]
# fmt: on


class TestBoundingVolumeBox(unittest.TestCase):
    @classmethod
    def build_box_sample(cls) -> BoundingVolumeBox:
        bounding_volume_box = BoundingVolumeBox()
        bounding_volume_box.set_from_list(DUMMY_MATRIX)
        return bounding_volume_box

    def test_constructor(self) -> None:
        bounding_volume_box = BoundingVolumeBox()
        self.assertIs(bounding_volume_box._box, None)

    def test_from_list(self) -> None:
        bounding_volume_box = BoundingVolumeBox.from_list(DUMMY_MATRIX)
        box = bounding_volume_box._box
        assert_array_equal(box, np.array(DUMMY_MATRIX))  # type: ignore [arg-type]

    def test_set_from_list(self) -> None:
        bounding_volume_box = BoundingVolumeBox()
        bounding_volume_box.set_from_list(DUMMY_MATRIX)
        box = bounding_volume_box._box
        assert_array_equal(box, np.array(DUMMY_MATRIX))  # type: ignore [arg-type]

    def test_set_from_invalid_list(self) -> None:
        bounding_volume_box = BoundingVolumeBox()

        # Empty list
        bounding_volume_list: List[float] = []
        with self.assertRaises(ValueError):
            bounding_volume_box.set_from_list(bounding_volume_list)
        self.assertIs(bounding_volume_box._box, None)

        # Too small list
        with self.assertRaises(ValueError):
            bounding_volume_box.set_from_list(DUMMY_MATRIX[:-1])
        self.assertIs(bounding_volume_box._box, None)

        # Too long list
        with self.assertRaises(ValueError):
            bounding_volume_box.set_from_list(DUMMY_MATRIX + [13])
        self.assertIs(bounding_volume_box._box, None)

        # Not only number
        with self.assertRaises(ValueError):
            bounding_volume_box.set_from_list(DUMMY_MATRIX[:-1] + ["a"])
        self.assertIs(bounding_volume_box._box, None)

        with self.assertRaises(ValueError):
            bounding_volume_box.set_from_list(DUMMY_MATRIX[:-1] + [[1]])
        self.assertIs(bounding_volume_box._box, None)

    def test_from_points(self) -> None:
        bounding_volume_box = BoundingVolumeBox.from_points(
            [np.array([1, 0, 0]), np.array([2, 0, 0])]
        )
        assert bounding_volume_box._box is not None
        assert_array_equal(
            bounding_volume_box._box, np.array([1.5, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0])
        )

    def test_set_from_points(self) -> None:
        bounding_volume_box = BoundingVolumeBox()
        bounding_volume_box.set_from_points([np.array([1, 0, 0]), np.array([2, 0, 0])])
        assert bounding_volume_box._box is not None
        assert_array_equal(
            bounding_volume_box._box, np.array([1.5, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0])
        )

        bounding_volume_box = BoundingVolumeBox()
        bounding_volume_box.set_from_points(
            [
                np.array([1, 0, 0]),
                np.array([2, 0, 0]),
                np.array([1, 1, 1]),
                np.array([2, 0, -1]),
            ]
        )
        assert bounding_volume_box._box is not None
        assert_array_equal(
            bounding_volume_box._box, [1.5, 0.5, 0, 0.5, 0, 0, 0, 0.5, 0, 0, 0, 1]
        )

    def test_set_from_invalid_points(self) -> None:
        # what if I give only one point ?
        pass

    def test_get_center(self) -> None:
        bounding_volume_box = BoundingVolumeBox()
        with self.assertRaises(AttributeError):
            bounding_volume_box.get_center()

        bounding_volume_box = TestBoundingVolumeBox.build_box_sample()
        assert_array_equal(bounding_volume_box.get_center(), [1, 2, 3])

    def test_translate(self) -> None:
        bounding_volume_box = BoundingVolumeBox()
        with self.assertRaises(AttributeError):
            bounding_volume_box.translate(np.array([-1, -2, -3]))

        bounding_volume_box = TestBoundingVolumeBox.build_box_sample()
        assert_array_equal(bounding_volume_box.get_center(), [1, 2, 3])

        bounding_volume_box.translate(np.array([-1, -2, -3]))
        # Should move only the center
        # fmt: off
        expected_result = [
            0, 0, 0, 4,
            5, 6, 7, 8,
            9, 10, 11, 12,
        ]
        # fmt: on
        box = bounding_volume_box._box
        assert_array_equal(box, expected_result)  # type: ignore [arg-type]

    def test_transform(self) -> None:
        bounding_volume_box = TestBoundingVolumeBox.build_box_sample()

        # Assert box hasn't change after transformation with identity matrix
        transformer = np.identity(4)
        bounding_volume_box.transform(transformer)
        assert_array_equal(bounding_volume_box._box, DUMMY_MATRIX)  # type: ignore [arg-type]

        # Assert box is translated by [10, 10, 10] on X,Y, Z axis
        transformer[:, 3] = 10
        bounding_volume_box.transform(transformer)
        # fmt: off
        expected_result = [
            11, 12, 13, 4,
            5, 6, 7, 8,
            9, 10, 11, 12,
        ]
        # fmt: on
        assert_array_equal(bounding_volume_box._box, expected_result)  # type: ignore [arg-type]

        # 90° rotation on z axis
        theta = math.pi / 2
        c, s = np.cos(theta), np.sin(theta)
        # fmt: off
        transformer = np.array(
            ((c, -s, 0, 0),
             (s, c, 0, 0),
             (0, 0, 1, 0),
             (0, 0, 0, 1))
        )
        # fmt: on
        bounding_volume_box = BoundingVolumeBox()
        bounding_volume_box.set_from_list([0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 0, 3])
        bounding_volume_box.transform(transformer)
        assert bounding_volume_box._box is not None
        # fmt: off
        assert_array_almost_equal(
            bounding_volume_box._box,
            [
                # same center
                0, 0, 0,
                # x,y half axis inverted
                0, 1, 0,
                -2, 0, 0,
                # same z half axis
                0, 0, 3,
            ],
        )
        # fmt: on

        # 45° rotation y axis
        theta = math.pi / 4
        c, s = np.cos(theta), np.sin(theta)
        transformer = np.array(
            # fmt: off
            ((c, 0, -s, 0),
             (0, 1, 0, 0),
             (s, 0, c, 0),
             (0, 0, 0, 1))
            # fmt: on
        )
        bounding_volume_box = BoundingVolumeBox()
        bounding_volume_box.set_from_list([0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 0, 3])
        bounding_volume_box.transform(transformer)
        assert bounding_volume_box._box is not None
        assert_array_almost_equal(
            bounding_volume_box._box,
            # fmt: off
            [
                # same center
                0, 0, 0,
                # x axis
                c, 0, s,
                # y unchanged
                0, 2, 0,
                -3 * c, 0, 3 * s,
            ],
            # fmt: on
        )

        # -30° deg rotation on x axis
        theta = -math.pi / 3
        c, s = np.cos(theta), np.sin(theta)
        # fmt: off
        transformer = np.array(
            ((1, 0, 0, 0),
             (0, c, -s, 0),
             (0, s, c, 0),
             (0, 0, 0, 1))
        )
        # fmt: on
        bounding_volume_box = BoundingVolumeBox()
        bounding_volume_box.set_from_list([0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 0, 3])
        bounding_volume_box.transform(transformer)
        assert bounding_volume_box._box is not None
        assert_array_almost_equal(
            bounding_volume_box._box,
            # fmt: off
            [
                # same center
                0, 0, 0,
                # x axis unchanged,
                1, 0, 0,
                0, 2 * c, 2 * s,
                0, -3 * s, 3 * c,
            ],
            # fmt: on
        )

    def test_get_corners(self) -> None:
        bounding_volume_box = BoundingVolumeBox()
        with self.assertRaises(AttributeError):
            bounding_volume_box.get_corners()

        bounding_volume_box = TestBoundingVolumeBox.build_box_sample()
        assert_array_equal(
            bounding_volume_box.get_corners(),
            [  # almost a kindness test
                [-20, -22, -24],
                [-12, -12, -12],
                [-6, -6, -6],
                [2, 4, 6],
                [0, 0, 0],
                [8, 10, 12],
                [14, 16, 18],
                [22, 26, 30],
            ],
        )

    def test_get_canonical_as_array(self) -> None:
        pass

    def test_to_dict(self) -> None:
        bounding_volume_box = BoundingVolumeBox()
        with self.assertRaises(AttributeError):
            bounding_volume_box.to_dict()

        self.assertDictEqual(
            TestBoundingVolumeBox.build_box_sample().to_dict(),
            {"box": DUMMY_MATRIX},
        )


if __name__ == "__main__":
    unittest.main()
